# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gromozeka/primitives/task.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='gromozeka/primitives/task.proto',
  package='protocol',
  syntax='proto3',
  serialized_pb=_b('\n\x1fgromozeka/primitives/task.proto\x12\x08protocol\";\n\x12ReplyToBrokerPoint\x12\x10\n\x08\x65xchange\x18\x01 \x01(\t\x12\x13\n\x0brouting_key\x18\x02 \x01(\t\"\xcc\x01\n\x04Task\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0f\n\x07task_id\x18\x02 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x03 \x01(\t\x12\x0e\n\x06kwargs\x18\x04 \x01(\t\x12\x0f\n\x07retries\x18\x05 \x01(\x05\x12\r\n\x05\x64\x65lay\x18\x06 \x01(\x04\x12\r\n\x05state\x18\x07 \x01(\t\x12\x14\n\x0c\x64\x65livery_tag\x18\x08 \x01(\x04\x12\x12\n\ngraph_uuid\x18\t \x01(\t\x12.\n\x08reply_to\x18\n \x01(\x0b\x32\x1c.protocol.ReplyToBrokerPointb\x06proto3')
)




_REPLYTOBROKERPOINT = _descriptor.Descriptor(
  name='ReplyToBrokerPoint',
  full_name='protocol.ReplyToBrokerPoint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='exchange', full_name='protocol.ReplyToBrokerPoint.exchange', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='routing_key', full_name='protocol.ReplyToBrokerPoint.routing_key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=45,
  serialized_end=104,
)


_TASK = _descriptor.Descriptor(
  name='Task',
  full_name='protocol.Task',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='protocol.Task.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='task_id', full_name='protocol.Task.task_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='args', full_name='protocol.Task.args', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kwargs', full_name='protocol.Task.kwargs', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='retries', full_name='protocol.Task.retries', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='delay', full_name='protocol.Task.delay', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='state', full_name='protocol.Task.state', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='delivery_tag', full_name='protocol.Task.delivery_tag', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='graph_uuid', full_name='protocol.Task.graph_uuid', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reply_to', full_name='protocol.Task.reply_to', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=107,
  serialized_end=311,
)

_TASK.fields_by_name['reply_to'].message_type = _REPLYTOBROKERPOINT
DESCRIPTOR.message_types_by_name['ReplyToBrokerPoint'] = _REPLYTOBROKERPOINT
DESCRIPTOR.message_types_by_name['Task'] = _TASK
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ReplyToBrokerPoint = _reflection.GeneratedProtocolMessageType('ReplyToBrokerPoint', (_message.Message,), dict(
  DESCRIPTOR = _REPLYTOBROKERPOINT,
  __module__ = 'gromozeka.primitives.task_pb2'
  # @@protoc_insertion_point(class_scope:protocol.ReplyToBrokerPoint)
  ))
_sym_db.RegisterMessage(ReplyToBrokerPoint)

Task = _reflection.GeneratedProtocolMessageType('Task', (_message.Message,), dict(
  DESCRIPTOR = _TASK,
  __module__ = 'gromozeka.primitives.task_pb2'
  # @@protoc_insertion_point(class_scope:protocol.Task)
  ))
_sym_db.RegisterMessage(Task)


# @@protoc_insertion_point(module_scope)



import functools
import json
import logging
import uuid
from collections import deque

import gromozeka
import gromozeka.concurrency
from gromozeka.exceptions import SerializationError
from gromozeka.primitives.dag import Vertex, Graph, GROUP, CHAIN, ACHAIN
from gromozeka.primitives.task_pb2 import Task as ProtoTask, ReplyToBrokerPoint

PENDING = 'PENDING'
RECEIVED = 'RECEIVED'
STARTED = 'STARTED'
SUCCESS = 'SUCCESS'
FAILURE = 'FAILURE'
CANCELED = 'CANCELED'
REJECTED = 'REJECTED'
RETRY = 'RETRY'


def task(bind=False, max_retries=0, retry_countdown=1, ignore_result=None, reply_to=None):
    """Task decorator. Use to make usual function as task

    After this, you must register task

    Examples:

        Register task:

        .. code-block:: python

            @task()
            def hello_world():
                print('hello world')

            app=Gromozeka()

            app.register(
                hello_world(),
                broker_point=BrokerPointType(
                    exchange="first_exchange",
                    exchange_type='direct',
                    queue='first_queue',
                    routing_key='first'))

            app.start()

        or :obj:`Task` object can register self:

        .. code-block:: python

            @task()
            def hello_world():
                print('hello world')

            app=Gromozeka()

            hello_world().register(
                broker_point=BrokerPointType(
                    exchange="first_exchange",
                    exchange_type='direct',
                    queue='first_queue',
                    routing_key='first'))

            app.start()

    Args:
        bind(:obj:`bool`, optional): If `True` `Task` will be in function arguments as first argument
        max_retries(:obj:`int`, optional): Maximum number of retries
        retry_countdown(:obj:`int` or :obj:`float`, optional): Pause between retries
        ignore_result(bool): If False result will be saved in result backend
        reply_to(BrokerPoint):

    Returns:
        gromozeka.primitives.Task: task object
    """

    def task_decor(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if reply_to:
                reply_to_exchange, reply_to_routing_key = reply_to.exchange, reply_to.routing_key
            else:
                reply_to_exchange, reply_to_routing_key = None, None
            return Task(func=f, bind=bind, max_retries=max_retries, retry_countdown=retry_countdown,
                        ignore_result=ignore_result, reply_to_exchange=reply_to_exchange,
                        reply_to_routing_key=reply_to_routing_key, args=args, kwargs=kwargs)

        return wrapper

    return task_decor


class RegistryTask:
    """Task registry class

    Attributes:
        task_id(:obj:`str`): Task identification
        func: Task function
        bind(:obj:`bool`): If `True` `Task` will be in function arguments as first argument
        broker_point(:obj:`gromozeka.primitives.BrokerPointType`): Broker entry
        worker_class(:obj:`gromozeka.concurrency.Worker`): Task workers will be this type
        max_workers(:obj:`int`): Number of task workers
        max_retries(:obj:`int`): Maximum number of retries
        retry_countdown(:obj:`int` or :obj:`float`): Pause between retries
        pool(:obj:`gromozeka.concurrency.Pool`): Task workers pool
    Args:
        broker_point(gromozeka.primitives.BrokerPointType): Broker entry
        task_id(str): Task identification
        func: Task function
        bind(bool): If `True` `Task` will be in function arguments as first argument
        worker_class(gromozeka.concurrency.Worker): Task workers will be this type
        max_workers(int): Number of task workers
        retry_countdown(:obj:`int` or :obj:`float`): Pause between retries
    """

    __slots__ = ['task_id', 'func', 'bind', 'broker_point', 'worker_class', 'max_workers', 'max_retries',
                 'retry_countdown', 'pool', 'ignore_result', 'deserializator']

    def __init__(self, broker_point, task_id, func, bind, worker_class, max_workers, max_retries, retry_countdown,
                 ignore_result):
        self.task_id = task_id
        self.func = func
        self.bind = bind
        self.broker_point = broker_point
        self.worker_class = worker_class or gromozeka.concurrency.ThreadWorker
        self.max_workers = max_workers
        self.max_retries = max_retries
        self.retry_countdown = retry_countdown
        self.ignore_result = ignore_result
        self.pool = gromozeka.concurrency.Pool(max_workers=self.max_workers, worker_class=self.worker_class)


class BrokerPoint:
    def __init__(self, exchange=None, exchange_type=None, queue=None, routing_key=None):
        self.exchange = exchange
        self.exchange_type = exchange_type
        self.queue = queue
        self.routing_key = routing_key

    def __str__(self):
        return '_'.join([v for v in to_dict(self).values() if v])


class TaskDeserializator:

    def deserialize(self, raw_task):
        """

        Args:
            raw_task: raw task to deserialize

        Returns:
            task properties, tuple task_id args kwargs uuid graph_uuid
        """
        raise NotImplementedError


class Task(Vertex):
    """Task class.

    Args:
        func: Task function
        args(:obj:`list`, optional):Task function arguments
        kwargs(:obj:`dict`, optional): Task function keyword arguments
        bind(bool): If `True` `Task` will be in function arguments as first argument
        state(:obj:`str`, optional): Current task state
        max_retries(:obj:`int`, optional): Maximum number of retries
        retry_countdown(:obj:`int` or :obj:`float`, optional): Pause between retries
        retries(:obj:`int`, optional): retries counter
        delay(:obj:`str`, optional): Retry countdown (%Y-%m-%d %H:%M:%S.%f)
        uuid_(:obj:`uuid.uuid4`, optional): Unique task runtime identification
        broker_point(gromozeka.primitives.BrokerPointType): Broker entry
        delivery_tag(:obj:`int`, optional): Original message delivery tag
    """

    __slots__ = ['uuid', 'task_id', 'logger', 'func', 'args', 'kwargs', 'bind', 'app', 'max_retries', 'retry_countdown',
                 'retries', 'delay', 'broker_point', 'delivery_tag', 'state', 'manual_ack',
                 'ignore_result', 'graph_uuid', 'reply_to_exchange', 'reply_to_routing_key']

    def __init__(self, func, args=None, kwargs=None, bind=None, app=None, state=None, max_retries=None,
                 retry_countdown=None, retries=None, delay=None, uuid_=None, broker_point=None, delivery_tag=None,
                 ignore_result=None, graph_uuid=None, reply_to_exchange=None, reply_to_routing_key=None):
        self.app = app or gromozeka.get_app()
        self.uuid = uuid_
        self.task_id = "%s.%s" % (func.__module__, func.__name__)
        self.logger = logging.getLogger("{}.pool.worker.{}".format(self.app.config.app_id, self.task_id))
        self.state = state or PENDING
        self.func = func
        self.args = args or []
        self.kwargs = kwargs or {}
        self.bind = bind or False
        self.max_retries = max_retries or 1
        self.retry_countdown = retry_countdown or 0
        self.retries = retries or 0
        self.delay = delay or 0
        self.broker_point = broker_point
        self.delivery_tag = delivery_tag or 0
        self.manual_ack = False
        self.ignore_result = ignore_result or False
        self.graph_uuid = graph_uuid
        self.error = None
        self.exc = None
        self.reply_to_exchange = reply_to_routing_key
        self.reply_to_routing_key = reply_to_exchange
        super().__init__(uuid=self.uuid, task_id=self.task_id, args=self.args, kwargs=self.kwargs, bind=self.bind,
                         reply_to_exchange=reply_to_exchange, reply_to_routing_key=reply_to_routing_key)

    def register(self, broker_point, worker_class=None, max_workers=1, max_retries=0, retry_countdown=0,
                 ignore_result=False, broker_options=None, deserializator=None):
        """Register task in task registry

        Args:
            broker_point(gromozeka.primitives.ProtoBrokerPoint): Broker entry
            worker_class(:class:`gromozeka.concurrency.Worker`): Task workers will be this type
            max_workers(int): Number of task workers
            max_retries(int): Maximum number of retries
            retry_countdown(:obj:`int` or :obj:`float`): Pause between retries
            ignore_result(bool): If False result will be saved in result backend
            broker_options: Specific broker options. See NatsOptions broker for example
            deserializator(TaskDeserializator): Custom task deserializator

        Returns:
            gromozeka.primitives.Task:
        """
        if deserializator:
            if not isinstance(deserializator, TaskDeserializator):
                raise SerializationError('custom deserializator should be inherited from `TaskDeserializator`')
        self.broker_point = broker_point
        r_task = RegistryTask(task_id=self.task_id, func=self.func, bind=self.bind, worker_class=worker_class,
                              max_workers=max_workers, max_retries=max_retries or self.max_retries,
                              retry_countdown=retry_countdown or self.retry_countdown or 1,
                              broker_point=broker_point or self.broker_point, ignore_result=ignore_result)
        self.app.register(task=r_task, broker_point=self.broker_point)
        self.app.broker_adapter.task_register(task_id=self.task_id, broker_point=broker_point, options=broker_options,
                                              deserializator=deserializator)
        return self

    def __call__(self, *args, **kwargs):
        """Execute `Task` function

        Examples:

            .. code-block:: python

                @task()
                def some_task(a,b):
                    print(a+b)

                some_task()(1,2)

        Args:
            *args: function arguments
            **kwargs: function keyword arguments

        Returns:
            function result
        """
        if self.bind:
            args[0:0] = [self]
        return functools.partial(self.func, *args, **kwargs)()

    @property
    def request(self):
        """Task `Request` property

        Returns:
            gromozeka.primitives.Request:
        """
        args = json.dumps(self.args) if self.args else None
        kwargs = json.dumps(self.kwargs) if self.kwargs else None
        if not self.broker_point:
            self.broker_point = self.app.get_task(self.task_id).broker_point
        if self.reply_to_exchange:
            reply_to = ReplyToBrokerPoint(exchange=self.reply_to_exchange,
                                          routing_key=self.reply_to_routing_key)
        else:
            reply_to = None
        return ProtoTask(uuid=self.uuid.__str__(), task_id=self.task_id, args=args, kwargs=kwargs, retries=self.retries,
                         delay=self.delay, state=self.state, delivery_tag=self.delivery_tag, graph_uuid=self.graph_uuid,
                         reply_to=reply_to)

    @classmethod
    def from_proto(cls, raw_task, delivery_tag=None, deserializator=False):
        """Make `Task` from `Request`

        Args:
            raw_task(gromozeka.primitives.Request): task `Request`
            delivery_tag(int): use with deserializator, this arg from broker
            deserializator(TaskDeserializator):

        Returns:
            gromozeka.primitives.Task:
        """
        app = gromozeka.get_app()
        if not deserializator:
            proto_task = ProtoTask()
            proto_task.ParseFromString(raw_task)
            task_uuid = proto_task.uuid
            graph_uuid = proto_task.graph_uuid
            task_id = proto_task.task_id
            args = None if not proto_task.args else json.loads(proto_task.args)
            kwargs = None if not proto_task.kwargs else json.loads(proto_task.kwargs)
            delay = proto_task.delay
            delivery_tag = proto_task.delivery_tag
            retries = proto_task.retries
            reply_to_exchange, reply_to_routing_key = proto_task.reply_to.exchange, proto_task.reply_to.routing_key

        else:
            task_uuid, task_id, graph_uuid, args, kwargs, \
            retries, delay, reply_to_exchange, reply_to_routing_key = deserializator.deserialize(raw_task=raw_task)
        r_task = app.get_task(task_id)
        return cls(func=r_task.func,
                   args=args,
                   kwargs=kwargs,
                   bind=r_task.bind,
                   app=app,
                   max_retries=r_task.max_retries,
                   retry_countdown=r_task.retry_countdown,
                   retries=retries,
                   delay=delay,
                   uuid_=task_uuid,
                   delivery_tag=delivery_tag,
                   ignore_result=r_task.ignore_result,
                   graph_uuid=graph_uuid,
                   broker_point=BrokerPoint(exchange=r_task.broker_point.exchange,
                                            exchange_type=r_task.broker_point.exchange_type,
                                            queue=r_task.broker_point.queue,
                                            routing_key=r_task.broker_point.routing_key),
                   reply_to_exchange=reply_to_exchange,
                   reply_to_routing_key=reply_to_routing_key)

    def apply_async(self, *args, **kwargs):
        """Run `Task` asynchronously

        Examples:

            .. code-block:: python

                @task()
                def some_task(a,b):
                    print(a+b)

                some_task().apply_async(1,2)

        Args:
            *args: `Task` function arguments
            **kwargs: `Task` function keyword arguments
        """
        self.uuid = self.uuid or str(uuid.uuid4())
        r_task = self.app.get_task(self.task_id)
        self.broker_point = r_task.broker_point
        self.args = args or self.args
        self.kwargs = kwargs or self.kwargs
        self.app.broker_adapter.task_send(task_uuid=self.uuid, request=self.request.SerializeToString(),
                                          broker_point=self.broker_point)

    def a(self, *args):
        self.args = args
        return self

    def k(self, **kwargs):
        self.kwargs = kwargs
        return self

    def on_receive(self):
        """Event when task received in worker

        """
        self.state = RECEIVED

    def on_start(self):
        """Event when started tu run task

        """
        self.state = STARTED

    def on_reject(self):
        """Event when task rejected

        """
        self.state = REJECTED
        self._reject()

    def on_cancel(self):
        """Event when task canceled

        """
        self.state = CANCELED
        self._reject()

    def on_retry(self, e):
        """Event when task on retry

        Args:
            e(gromozeka.exceptions.Retry):

        Returns:
             bool: `True` if task will retry more times
        """

        self.state = RETRY
        if not e.max_retries:
            e.max_retries = self.max_retries
        if not e.retry_countdown:
            e.retry_countdown = self.retry_countdown
        if e.max_retries <= self.retries:
            return False
        if self.graph_uuid:
            graph_dict = self.app.backend_adapter.graph_get(self.graph_uuid)
            graph = Graph.from_dict(graph_dict)
            current = graph.vertex_by_uuid(self.uuid)
            graph.state = current.state = self.state
            current.error = self.error
            self.app.backend_adapter.graph_update(graph_uuid=graph.uuid,
                                                  verticies=[current.as_dict()],
                                                  graph_state=graph.state, error_task_uuid=current.uuid,
                                                  short_error=self.error)
        self._retry(e)
        return True

    def on_success(self, res):
        """Event when task finishes successfully

        """
        if self.state == CANCELED:
            self._reject()
            return

        self.state = SUCCESS
        self.retries = 0

        # result save
        if not self.graph_uuid:
            if not self.ignore_result:
                self.app.backend_adapter.result_set(self.uuid, res)
            self._ack()
            return

        if not self.ignore_result:
            self.app.backend_adapter.result_set(self.uuid, res, graph_uuid=self.graph_uuid)

        graph_dict = self.app.backend_adapter.graph_get(self.graph_uuid)
        graph = Graph.from_dict(graph_dict)
        current = graph.vertex_by_uuid(self.uuid)
        current.state = self.state
        self.update_graph_states(graph, current, result=res)
        for v in graph.next_vertex(from_uuid=current.uuid):
            if v.args_from:
                if v.args_from == current.fr:
                    args_from = res
                elif graph.vertex_by_uuid(v.args_from).operation == GROUP:
                    args_from = self.app.backend_adapter.group_get_result(graph_uuid=graph.uuid, group_uuid=v.args_from)
                else:
                    args_from_parent_vertex = graph.vertex_by_uuid(v.args_from)
                    args_from_vertex = graph.vertex_by_uuid(args_from_parent_vertex.children[1]).uuid
                    args_from = self.app.backend_adapter.result_get(graph_uuid=graph.uuid,
                                                                    task_uuid=args_from_vertex)
                if v.bind:
                    v.args[0:0] = [args_from] if v.bind else [args_from] + v.args
                else:
                    v.args = [args_from] + v.args
            v.apply(graph.uuid)

        # acknowledge broker message
        self._ack()

    def update_graph_states(self, graph, vertex, graph_state=None, result=None, error=None):
        """

        Args:
            graph(gromozeka.primitives.dag.Graph): graph instance
            vertex(gromozeka.primitives.dag.Vertex): vertex instance
            graph_state:
            result: Graph vertex result
            error(str): error string

        Returns:
            list of gromozeka.primitives.dag.Vertex:
        """
        stack = deque([vertex])
        completed = []
        error_task_uuid = vertex.uuid if error else None
        while stack:
            current = stack[-1]
            op = current.operation
            if not op:
                current.state = vertex.state
                parent = graph.vertex_by_uuid(current.parent)
                if parent:
                    if parent.operation in (CHAIN, ACHAIN):
                        if graph.is_chain_tail(current):
                            stack.appendleft(parent)
                    else:
                        stack.appendleft(parent)
            elif op == GROUP:
                group_len = self.app.backend_adapter.group_add_result(
                    graph_uuid=graph.uuid,
                    group_uuid=current.uuid,
                    task_uuid=vertex.uuid,
                    result=result if not self.ignore_result else None)
                if group_len != len(current.children):
                    stack.pop()
                    continue
                current.state = vertex.state
                parent = graph.vertex_by_uuid(current.parent)
                if parent:
                    if parent.operation in (CHAIN, ACHAIN):
                        if graph.is_chain_tail(current):
                            stack.appendleft(parent)
                    else:
                        stack.appendleft(parent)
            elif current.state == vertex.state:
                if current.parent:
                    stack.appendleft(graph.vertex_by_uuid(current.parent))
            elif graph.vertex_by_uuid(current.children[1]) in completed:
                graph.vertex_by_uuid(current.children[0]).state = vertex.state
                current.state = vertex.state
                if current.parent:
                    stack.appendleft(graph.vertex_by_uuid(current.parent))
            else:
                stack.pop()
                continue
            if current.uuid == graph.last_uuid:
                graph_state = vertex.state
            completed.append(current)
            stack.pop()
        self.app.backend_adapter.graph_update(graph_uuid=graph.uuid, verticies=[comp.as_dict() for comp in completed],
                                              graph_state=graph_state, error_task_uuid=error_task_uuid,
                                              short_error=error)

    def on_fail(self):
        """Event when task failed with `Exception`

        """
        self.state = FAILURE
        if self.graph_uuid:
            graph_dict = self.app.backend_adapter.graph_get(self.graph_uuid)
            graph = Graph.from_dict(graph_dict)
            current = graph.vertex_by_uuid(self.uuid)
            current.state = self.state
            current.error = self.error
            current.exc = self.exc
            self.update_graph_states(graph=graph, vertex=current, graph_state=self.state, error=self.error)
        self._reject()

    def cancel(self):
        """Cancel task. Use this method in `Task` function. Bind option must be `True`

        Examples:
            .. code-block:: python

                @task(bind=True)
                def some_task():
                    try:
                        a = 1 / 0
                    except ZeroDivisionError as e:
                        self.cancel()
        """
        self.on_cancel()

    def _retry(self, e):
        """

        Args:
            e(gromozeka.exceptions.Retry):

        """
        self.logger.warning("retry on error: {0}".format(e))
        self.delay = e.retry_countdown
        self.retries += 1
        self.app.broker_adapter.task_send_delayed(task_uuid=self.uuid, request=self.request.SerializeToString(),
                                                  broker_point=self.broker_point, delay=self.delay)
        self._reject()

    def _ack(self):
        if self.manual_ack:
            return
        self.app.broker_adapter.task_done(task_uuid=self.uuid, broker_point=self.broker_point,
                                          delivery_tag=self.delivery_tag)

    def _result_set(self, result):
        return self.app.backend_adapter.result_set(self.uuid, result)

    def _reject(self):
        self.app.broker_adapter.task_reject(task_uuid=self.uuid, broker_point=self.broker_point,
                                            delivery_tag=self.delivery_tag)


def to_dict(o):
    return {k: getattr(o, k) for k in o.__dir__() if
            not callable(getattr(o, k)) and not k.startswith('__')}



"""Module for schedule tasks

"""
import logging
import queue
import sched
import threading

import time
from datetime import datetime, timedelta


class Scheduler(threading.Thread):
    """Scheduler class

    Attributes:
        logger(:obj:`logging.Logger`): Class logger
        _stop_event(:obj:`threading.Event`): Stop event
        _shed(:obj:`sched.scheduler`) : Scheduler instance
        _event_queue(:obj:`queue.Queue`): Queue for scheduler events
    """

    __slots__ = ['logger', '_shed', '_event_queue', '_stop_event']

    def __init__(self):
        self.logger = logging.getLogger('gromozeka.scheduler')
        self._shed = sched.scheduler(time.time, time.sleep)
        self._event_queue = queue.Queue()
        self._stop_event = threading.Event()
        super().__init__()

    def run(self):
        """Run scheduler in thread

        """
        self.logger.info("start")
        self.listen_cmd()

    def add(self, wait_time, func, args=None, kwargs=None, priority=1):
        """Add event to scheduler queue

        Args:
            wait_time(float): Delay to run (seconds)
            func: Callable function to run
            args: Func arguments
            kwargs: Func keyword arguments
            priority: Event priority
        """
        self._event_queue.put((wait_time, priority, func, args or [], kwargs or {}))

    def stop(self):
        """Stop scheduler

        """
        self._stop_event.set()

    def listen_cmd(self):
        """Listen to commands

        """
        while not self._stop_event.is_set():
            self._shed.run(blocking=False)
            try:
                event = self._event_queue.get(True, 0.05)
            except queue.Empty:
                continue
            self._shed.enter(delay=event[0], priority=event[1], action=event[2], argument=event[3], kwargs=event[4])
        self.logger.info("stop")


def delay_from_eta(eta):
    """Make new delay from eta

    Examples:

        .. code-block:: python

            delay_from_eta(countdown='2018-02-20 13:38:33.0') # This event will run 20.02.18 at 13:38:33 once

    Args:
        eta(:obj:`str`): If `exceptions.Retry` occurs other retry will run at this time.
         Other arguments ignored.

    Returns:
        :obj:`int` or :obj:`None`: Delay
    """
    try:
        return time.mktime(time.strptime(eta, "%Y-%m-%d %H:%M:%S.%f")) - time.time()
    except ValueError:
        return time.mktime(time.strptime(eta, "%Y-%m-%d %H:%M:%S")) - time.time()


def new_eta(seconds):
    """Make new eta from summary of current time and `seconds`

    Args:
        seconds(float): Seconds

    Returns:
        str: Datetime in format "%Y-%m-%d %H:%M:%S.%f"
    """
    return (datetime.now() + timedelta(seconds=seconds)).strftime("%Y-%m-%d %H:%M:%S.%f")


def delay(last_run=None, every=None, interval=None, at=None):
    """Make new delay from arguments

    Examples:

        .. code-block:: python

            delay(every='days', interval=2) # This event will run every 2 days

            delay(every='weeks', interval=1, at='15:40') # This event will run every week at 15:40

    Args:
        last_run(:obj:`str`, optional): Last run time
        every(:obj:`str`, optional): May be seconds, minutes, hours, days, weeks.
        interval(:obj:`int`, optional): Uses with `every`.
        at(:obj:`str`, optional): Time formatted as `hour:minute` or `hour:minute:second`.

    Returns:
        :obj:`int` or False: new event delay. Returns False when not need more event to run - last run.
    """
    if every:
        if last_run:
            last_run_timestamp = time.mktime(time.strptime(last_run, "%Y-%m-%d %H:%M:%S.%f"))
            interval_seconds = timedelta(**{every: (interval or 1)}).total_seconds()
            next_timestamp = last_run_timestamp + interval_seconds
            next_run_date = datetime.fromtimestamp(next_timestamp)
        else:
            next_run_date = datetime.now()
        if at and every in ('days', 'weeks'):
            time_list = at.split(":")
            hour, minute, second = map(lambda x: int(x),
                                       (time_list[0], time_list[1], time_list[2]) if len(time_list) == 3 else (
                                           time_list[0], time_list[1], 0))
            next_run_date = next_run_date.replace(hour=hour, minute=minute, second=second)
            while next_run_date < datetime.now() or next_run_date.timestamp() < 0:
                next_run_date += timedelta(days=1)
        return (next_run_date - datetime.now()).total_seconds()



import asyncio
import multiprocessing
import threading

from gromozeka.brokers.base import BrokerAdapter, get_broker_factory
from gromozeka.brokers.rabbit import RabbitMQPikaAdaptee


class LatexSema:
    def __init__(self, sema_cls, value):
        self.sema = sema_cls(value)
        if isinstance(self.sema, asyncio.Semaphore):
            self.lock = asyncio.Lock()
        elif isinstance(self.sema, threading.Semaphore):
            self.lock = threading.Lock()
        elif isinstance(self.sema, multiprocessing.Semaphore):
            self.lock = multiprocessing.Lock()
        else:
            raise Exception("unknown semaphore class")

    def acquire(self):
        self.sema.acquire()

    def change(self, delta):
        with self.lock:
            if delta > 0:
                for _ in range(delta):
                    self.release()
            elif delta < 0:
                self.sema._value += delta

    def release(self):
        if self.sema._value >= 0:
            self.sema.release()
        else:
            self.sema._value += 1


#
# class AsyncioSemaphore(asyncio.Semaphore):
#     def change(self, delta):
#         if delta > 0:
#             for _ in range(delta):
#                 self.release()
#         elif delta < 0:
#             self._value += delta
#
#     def release(self):
#         if self._value >= 0:
#             super().release()
#         else:
#             self._value += 1


__all__ = ['BrokerAdapter', 'RabbitMQPikaAdaptee', 'get_broker_factory', 'LatexSema']



from gromozeka.backends.base import BackendAdapter, get_backend_factory
from gromozeka.backends.redis import RedisAioredisAdaptee

__all__ = ['BackendAdapter', 'RedisAioredisAdaptee', 'get_backend_factory']



import logging
import multiprocessing
import os
import signal
import sys

from gromozeka.backends import get_backend_factory
from gromozeka.brokers import get_broker_factory
from gromozeka.concurrency import Scheduler
from gromozeka.config import Config
from gromozeka.exceptions import GromozekaException
from gromozeka.primitives import RegistryTask


def get_app():
    """

    Returns:
         gromozeka.app.Gromozeka:
    """
    try:
        return app
    except NameError:
        logging.getLogger("gromozeka").error("Gromozeka's  application not initialized")
        sys.exit(1)


class Gromozeka:
    """

    Attributes:
        logger(:obj:`logging.Logger`): Class logger
        registry(:obj:`dict` of :obj:`gromozeka.primitives.Task`): Task registry
        config(:obj:`gromozeka.config.Config`): config object
        _broker_adapter(:obj:`gromozeka.brokers.BrokerAdapter`): Broker
        _backend_adapter(:obj:`gromozeka.backends.BackendAdapter`): Broker
        _scheduler(:obj:`gromozeka.concurrency.Scheduler`): Scheduler
    """

    __slots__ = ['logger', '_broker_adapter', '_backend_adapter', '_scheduler', 'registry', 'config', 'pid',
                 '_exit_signum', 'is_closing']

    def __init__(self):
        global app
        app = self
        self.config = Config()
        self.logger = logging.getLogger('gromozeka')
        self._broker_adapter = None
        self._backend_adapter = None
        self._scheduler = None
        self.registry = {}
        self.pid = multiprocessing.current_process().pid
        self._exit_signum = 0
        self.is_closing = False

    @property
    def broker_adapter(self):
        if not self._broker_adapter:
            self._broker_adapter = get_broker_factory(self)
        return self._broker_adapter

    @property
    def backend_adapter(self):
        if not self._backend_adapter:
            self._backend_adapter = get_backend_factory(self)
        return self._backend_adapter

    @property
    def scheduler(self):
        if not self._scheduler:
            self._scheduler = Scheduler()
        return self._scheduler

    def config_from_env(self):
        """Configure Gromozeka with environment variables

        Examples:
            .. code-block:: python

                app=Gromozeka().config_from_env()

        Returns:
            gromozeka.Gromozeka: Configured application
        """
        self.config.from_env()
        return self

    def config_from_dict(self, conf):
        """Configure Gromozeka with :obj:`dict`

        Examples:
            .. code-block:: python

                conf={'app_prefix':'my_application','broker_reconnect_max_retries':3}
                app=Gromozeka().config_from_dict(conf)

        Args:
            conf(dict): config :obj:`dict`

        Returns:
            gromozeka.Gromozeka: Configured application
        """
        self.config.from_dict(conf)
        return self

    def start(self):
        """Start application

        """
        # setup signal handler
        for sig in ('SIGTERM', 'SIGINT'):
            signal.signal(getattr(signal, sig), self._signal_handler)
        self.backend_adapter.start()
        for _, registry_task in self.registry.items():
            if not hasattr(registry_task, 'pool'):
                continue
            registry_task.pool.start()
        self.broker_adapter.start()

    def register(self, task, broker_point, worker_class=None, max_workers=1, max_retries=0, retry_countdown=0,
                 ignore_result=False, broker_options=None, deserializator=None):
        """

        Args:
            task(:obj:`gromozeka.primitives.Task` or :obj:`gromozeka.primitives.RegistryTask`): Task
            broker_point(gromozeka.primitives.BrokerPointType): Broker entry
            worker_class(:class:`gromozeka.concurrency.Worker`, optional): Worker class
            max_workers(int): How match workers will start
            max_retries(int): Maximum number of retries, after it will reached task will down
            retry_countdown(int): Pause between retries (seconds)
            ignore_result(bool): If True result will be saved in result backend
            broker_options: Specific broker options. See NatsOptions broker for example
            deserializator(gromozeka.primitives.base.TaskDeserializator):
        """
        if isinstance(task, RegistryTask):
            self.registry[task.task_id] = task
        else:
            task.register(broker_point=broker_point, worker_class=worker_class, max_workers=max_workers,
                          max_retries=max_retries, retry_countdown=retry_countdown,
                          ignore_result=task.ignore_result or ignore_result,
                          broker_options=broker_options, deserializator=deserializator)

    def get_task(self, id_):
        """Get task by id

        Args:
            id_(str): Unique task identification

        Returns:
            gromozeka.primitives.RegistryTask: registry task
        """
        try:
            return self.registry[id_]
        except KeyError:
            raise GromozekaException('task `{}` not registered in registry'.format(id_))

    def _signal_handler(self, signum, _):
        """Signal handler

        """
        if multiprocessing.current_process().pid == self.pid:
            self.stop()
            if self._exit_signum:
                os._exit(signum)

    def stop_signal(self, signum):
        self._exit_signum = signum
        os.kill(self.pid, signum)

    def stop(self):
        self.logger.info('trying graceful shutdown')
        self.is_closing = True
        [rt.pool.stop() for rt in self.registry.values() if hasattr(rt, 'pool')]
        self.backend_adapter.stop()
        self.broker_adapter.stop()
        os._exit(0)



import asyncio
import json
import signal
from functools import wraps

import aioredis
import uvloop
from aioredis import ConnectionClosedError, ReplyError
from redis import ConnectionError
from redlock import RedLock

from gromozeka.backends.base import BackendInterface

REDIS_CONNECTION_EXCEPTIONS = (
    OSError, ConnectionClosedError, ReplyError, aioredis.errors.PipelineError, ConnectionError)

TASK_RESULT_PATTERN = 'gromozeka:task:{task_uuid}'  # hash
GRAPH_META_PATTERN = 'gromozeka:graph:{graph_uuid}'  # hash
GRAPH_LOCK_PATTERN = 'gromozeka:graph:{graph_uuid}:lock'  # redlock
GRAPH_TASK_PATTERN = 'gromozeka:graph:{graph_uuid}:task:{task_uuid}'  # key
GRAPH_TASK_GROUP_RESULT_PATTERN = 'gromozeka:graph:{graph_uuid}:group:{group_uuid}:result'  # list
GRAPH_TASK_GROUP_UUIDS_PATTERN = 'gromozeka:graph:{graph_uuid}:group:{group_uuid}:uuids'  # list
GRAPH_TASK_GROUP_LOCK_PATTERN = 'gromozeka:graph:{graph_uuid}:group:{group_uuid}:lock'  # redlock
GRAPH_TASK_CHAIN_RESULT_PATTERN = 'gromozeka:graph:{graph_uuid}:chain:{chain_uuid}'  # key


def async_error_handler(func):
    """

    Returns:
        function:
    """

    @wraps(func)
    async def wrap(*args, **kwargs):
        self = args[0]
        if 'retries' in kwargs:
            retries = kwargs['retries']
            del (kwargs['retries'])
        else:
            retries = self.reconnect_max_retries
        try:
            return await func(*args, **kwargs)
        except REDIS_CONNECTION_EXCEPTIONS as e:
            if isinstance(e, OSError) and 'Errno 61' not in str(e):
                raise
            if not retries:
                self.app.stop_signal(signal.SIGTERM)
            retries -= 1
            await asyncio.sleep(self.reconnect_retry_countdown)
            self.logger.info('trying to reconnect on error: %s', e)
            await wrap(*args, **kwargs, retries=retries)

    return wrap


class RedisAioredisAdaptee(BackendInterface):
    result_queue = None
    reconnect_max_retries = None
    reconnect_retry_countdown = None
    _connection = None
    _loop = None
    _url = None

    def __init__(self, app):
        """

        Args:
            app(gromozeka.Gromozeka):
        """
        self.app = app
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        super().__init__(app=app)

    def configure(self):
        self._url = self.app.config.backend_url
        self.reconnect_max_retries = self.app.config.backend_reconnect_max_retries
        self.reconnect_retry_countdown = self.app.config.backend_reconnect_retry_countdown

    @staticmethod
    def worker_run(self):
        """

        Args:
            self(gromozeka.backends.base.BackendAdapter):
        """
        self.backend._loop = asyncio.new_event_loop()
        self.backend._loop.create_task(self.backend.serve())
        self.backend._loop.run_forever()

    @async_error_handler
    async def serve(self):
        """Start `Backend`

        """
        self._connection = await aioredis.create_redis_pool(self._url, encoding='utf-8')

    def stop(self):
        if self._connection:
            asyncio.run_coroutine_threadsafe(self._stop(), loop=self._loop).result()
        asyncio.run_coroutine_threadsafe(self._loop.shutdown_asyncgens(), loop=self._loop)
        self._loop.stop()

    def result_get(self, task_uuid, graph_uuid=None):
        self.result_queue.put(
            asyncio.run_coroutine_threadsafe(self._result_get(task_uuid, graph_uuid), loop=self._loop).result())

    def result_set(self, task_uuid, result, graph_uuid=None):
        asyncio.run_coroutine_threadsafe(self._result_set(task_uuid=task_uuid, result=result, graph_uuid=graph_uuid),
                                         loop=self._loop).result()
        self.result_queue.put(True)

    def result_del(self, task_uuid):
        raise NotImplementedError

    def results_del(self, *task_uuids):
        raise NotImplementedError

    def graph_init(self, graph_dict):
        self.result_queue.put(asyncio.run_coroutine_threadsafe(self._graph_init(
            graph_dict=graph_dict), loop=self._loop).result())

    def graph_get(self, graph_uuid):
        r = asyncio.run_coroutine_threadsafe(self._graph_get(graph_uuid=graph_uuid), loop=self._loop).result()
        self.result_queue.put(r)

    def graph_update(self, graph_uuid, verticies, graph_state=None, error_task_uuid=None, short_error=None):
        res = asyncio.run_coroutine_threadsafe(
            self._graph_update(graph_uuid=graph_uuid, verticies=verticies, graph_state=graph_state,
                               error_task_uuid=error_task_uuid, short_error=short_error),
            loop=self._loop).result()
        self.result_queue.put(res)

    def graph_result_set(self, graph_uuid, task_uuid, result):
        res = asyncio.run_coroutine_threadsafe(
            self._graph_result_set(graph_uuid=graph_uuid, task_uuid=task_uuid, result=result),
            loop=self._loop).result()
        self.result_queue.put(res)

    def is_group_completed(self, graph_uuid, group_uuid, expected):
        res = asyncio.run_coroutine_threadsafe(
            self._is_group_completed(graph_uuid=graph_uuid, group_uuid=group_uuid, expected=expected),
            loop=self._loop).result()
        self.result_queue.put(res)

    def group_add_result(self, graph_uuid, group_uuid, task_uuid, result):

        res = asyncio.run_coroutine_threadsafe(
            self._group_add_result(graph_uuid=graph_uuid, group_uuid=group_uuid, task_uuid=task_uuid,
                                   result=result),
            loop=self._loop).result()
        self.result_queue.put(res)

    def group_get_result(self, graph_uuid, group_uuid):
        self.result_queue.put(
            asyncio.run_coroutine_threadsafe(self._group_get_result(graph_uuid=graph_uuid, group_uuid=group_uuid),
                                             loop=self._loop).result())

    def chain_get_result(self, graph_uuid, chain_uuid):
        self.result_queue.put(
            asyncio.run_coroutine_threadsafe(self._chain_get_result(graph_uuid=graph_uuid, chain_uuid=chain_uuid),
                                             loop=self._loop).result())

    @async_error_handler
    async def _stop(self):
        self._connection.close()
        await self._connection.wait_closed()

    @async_error_handler
    async def _result_get(self, task_uuid, graph_uuid=None):
        res = await self._connection.get(
            TASK_RESULT_PATTERN.format(task_uuid=task_uuid) if not graph_uuid else GRAPH_TASK_PATTERN.format(
                graph_uuid=graph_uuid, task_uuid=task_uuid))
        return json.loads(res)

    @async_error_handler
    async def _result_set(self, task_uuid, result, graph_uuid=None):
        task_key = TASK_RESULT_PATTERN.format(task_uuid=task_uuid) if not graph_uuid else GRAPH_TASK_PATTERN.format(
            graph_uuid=graph_uuid, task_uuid=task_uuid)
        pipe = self._connection.pipeline()
        pipe.set(task_key, json.dumps(result))
        pipe.expire(task_key, self.app.config.backend_results_timelife)
        return await pipe.execute()

    @async_error_handler
    async def _result_del(self, task_uuid):
        raise NotImplemented

    @async_error_handler
    async def _results_del(self, *task_uuids):
        raise NotImplemented

    @async_error_handler
    async def _graph_init(self, graph_dict):
        graph_hash = {
            'uuid': graph_dict['uuid'],
            'state': graph_dict['state'],
            'first_uuid': graph_dict['first_uuid'],
            'last_uuid': graph_dict['last_uuid'],
            'error_task_uuid': '',
            'short_error': '',
        }
        for k, v in graph_dict['verticies'].items():
            graph_hash[k] = json.dumps(v)
        graph_ident = GRAPH_META_PATTERN.format(graph_uuid=graph_dict['uuid'])
        pipe = self._connection.pipeline()
        pipe.hmset_dict(graph_ident, graph_hash)
        pipe.expire(graph_ident, self.app.config.backend_results_timelife)
        return await pipe.execute()

    @async_error_handler
    async def _graph_get(self, graph_uuid):
        return await self._connection.hgetall(GRAPH_META_PATTERN.format(graph_uuid=graph_uuid))

    @async_error_handler
    async def _graph_update(self, graph_uuid, verticies, graph_state=None, error_task_uuid=None, short_error=None):
        pipe = self._connection.pipeline()
        graph_ident = GRAPH_META_PATTERN.format(graph_uuid=graph_uuid)
        for vertex in verticies:
            pipe.hset(graph_ident, vertex['uuid'], json.dumps(vertex))
        if graph_state:
            pipe.hset(graph_ident, 'state', graph_state)
            pipe.hset(graph_ident, 'error_task_uuid', error_task_uuid or '')
            pipe.hset(graph_ident, 'short_error', short_error or '')
        await pipe.execute()

    @async_error_handler
    async def _graph_result_set(self, graph_uuid, task_uuid, result):
        await self._connection.set(GRAPH_TASK_PATTERN.format(graph_uuid=graph_uuid, task_uuid=task_uuid),
                                   json.dumps(result))

    @async_error_handler
    async def _is_group_completed(self, graph_uuid, group_uuid, expected):
        return (await self._connection.llen(
            GRAPH_TASK_GROUP_UUIDS_PATTERN.format(graph_uuid=graph_uuid, group_uuid=group_uuid))) >= expected

    @async_error_handler
    async def _group_add_result(self, graph_uuid, group_uuid, task_uuid, result):
        with RedLock('%s_%s' % (graph_uuid, group_uuid), connection_details=[{'url': self._url}]):
            uuids_key = GRAPH_TASK_GROUP_UUIDS_PATTERN.format(graph_uuid=graph_uuid, group_uuid=group_uuid)
            results_key = GRAPH_TASK_GROUP_RESULT_PATTERN.format(graph_uuid=graph_uuid, group_uuid=group_uuid)
            pipe = self._connection.pipeline()
            pipe.rpush(uuids_key, task_uuid)
            pipe.rpush(results_key, json.dumps(result))
            pipe.expire(uuids_key, self.app.config.backend_results_timelife)
            pipe.expire(results_key, self.app.config.backend_results_timelife)
            group_len = pipe.llen(uuids_key)
            await pipe.execute()
            return await group_len

    @async_error_handler
    async def _group_get_result(self, graph_uuid, group_uuid):
        res = await self._connection.lrange(
            GRAPH_TASK_GROUP_RESULT_PATTERN.format(graph_uuid=graph_uuid, group_uuid=group_uuid), 0, -1)
        return [json.loads(r) for r in res]

    @async_error_handler
    async def _chain_get_result(self, graph_uuid, chain_uuid):
        return await self._connection.get(
            GRAPH_TASK_CHAIN_RESULT_PATTERN.format(graph_uuid=graph_uuid, chain_uuid=chain_uuid))



import logging
import queue

from gromozeka.concurrency import Pool, commands
from gromozeka.concurrency import ThreadWorker
from gromozeka.exceptions import GromozekaException


def get_backend_factory(app):
    """

    Args:
        app (gromozeka.app.app.Gromozeka):

    Returns:
        gromozeka.backends.base.BackendInterface:
    """
    if app.config.backend_url.startswith('redis://'):
        from gromozeka.backends import RedisAioredisAdaptee
        backend_adapter = BackendAdapter(RedisAioredisAdaptee(app=app))
    else:
        raise GromozekaException('unknown backend connection url')
    backend_adapter.backend.configure()
    return backend_adapter


class BackendInterface:
    def __init__(self, app):
        """

        Args:
            app(gromozeka.Gromozeka):
        """
        self.logger = logging.getLogger("gromozeka.backend")
        self.app = app

    @staticmethod
    def worker_run(self):
        """This method will be in thread - start to work here

        """
        raise NotImplementedError

    def stop(self):
        """Stop backend_adapter

        """
        raise NotImplementedError

    def configure(self):
        raise NotImplementedError

    def result_get(self, task_uuid, graph_uuid=None):
        raise NotImplementedError

    def result_set(self, task_uuid, result, graph_uuid=None):
        raise NotImplementedError

    def result_del(self, task_uuid):
        raise NotImplementedError

    def results_del(self, *task_uuids):
        raise NotImplementedError

    def graph_init(self, graph_dict):
        raise NotImplementedError

    def graph_get(self, graph_uuid):
        raise NotImplementedError

    def graph_update(self, graph_uuid, verticies, graph_state=None, error_task_uuid=None, short_error=None):
        raise NotImplementedError

    def graph_result_set(self, graph_uuid, task_uuid, result):
        raise NotImplementedError

    def is_group_completed(self, graph_uuid, group_uuid, expected):
        raise NotImplementedError

    def group_add_result(self, graph_uuid, group_uuid, task_uuid, result):
        raise NotImplementedError

    def group_get_result(self, graph_uuid, group_uuid):
        raise NotImplementedError

    def chain_get_result(self, graph_uuid, chain_uuid):
        raise NotImplementedError


class BackendAdapter(Pool):
    def __init__(self, backend, logger=None):
        """

        Args:
            backend (gromozeka.backends.redis.RedisAioredisAdaptee):
            logger (logging.Logger):
        """
        self.logger = logger or logging.getLogger("gromozeka.backend")
        self.backend = backend
        self.worker = type('BackendWorker', (ThreadWorker,),
                           {'app': self.backend.app, 'backend': self.backend, 'run': self.backend.worker_run,
                            'logger': self.logger})
        super().__init__(max_workers=1, worker_class=self.worker, logger=self.logger)
        self.backend.result_queue = self.cmd_out

    def listen_cmd(self):
        """Listen to commands

        """
        while not self._stop_event.is_set():
            try:
                command, args = self.cmd.get(0.05)
            except queue.Empty:
                continue
            if command == commands.POOL_GROW:
                super()._grow(args)
            elif command == commands.POOL_STOP:
                self.backend.stop()
                super().stop_()
            elif command == commands.BACKEND_RESULT_GET:
                task_uuid, graph_uuid = args['task_uuid'], args['graph_uuid']
                self.backend.result_get(task_uuid=task_uuid, graph_uuid=graph_uuid)
            elif command == commands.BACKEND_RESULT_SET:
                task_uuid, result, graph_uuid = args['task_uuid'], args['result'], args['graph_uuid']
                self.backend.result_set(task_uuid=task_uuid, result=result, graph_uuid=graph_uuid)
                self.logger.info('result saved, task_uuid <%s>', task_uuid)
            elif command == commands.BACKEND_RESULT_DEL:
                self.backend.result_del(task_uuid=args['task_uuid'])
            elif command == commands.BACKEND_RESULT_DEL:
                self.backend.results_del(*args)
            elif command == commands.BACKEND_GRAPH_INIT:
                self.backend.graph_init(graph_dict=args)
            elif command == commands.BACKEND_GRAPH_GET:
                self.backend.graph_get(graph_uuid=args)
            elif command == commands.BACKEND_GRAPH_UPDATE:
                graph_uuid, verticies, graph_state = args['graph_uuid'], args['verticies'], args['graph_state']
                error_task_uuid, short_error = args['error_task_uuid'], args['short_error']
                self.backend.graph_update(graph_uuid=graph_uuid, verticies=verticies, graph_state=graph_state,
                                          error_task_uuid=error_task_uuid, short_error=short_error)
            elif command == commands.BACKEND_GRAPH_RESULT_SET:
                graph_uuid, task_uuid, result = args['graph_uuid'], args['task_uuid'], args['result']
                self.backend.graph_result_set(graph_uuid=graph_uuid, task_uuid=task_uuid, result=result)
            elif command == commands.BACKEND_IS_GROUP_COMPLETED:
                graph_uuid, group_uuid, expected = args['graph_uuid'], args['group_uuid'], args['expected']
                self.backend.is_group_completed(graph_uuid=graph_uuid, group_uuid=group_uuid, expected=expected)
            elif command == commands.BACKEND_GROUP_ADD_RESULT:
                graph_uuid, group_uuid = args['graph_uuid'], args['group_uuid']
                task_uuid, result = args['task_uuid'], args['result']
                self.backend.group_add_result(graph_uuid=graph_uuid, group_uuid=group_uuid, task_uuid=task_uuid,
                                              result=result)
            elif command == commands.BACKEND_GROUP_GET_RESULT:
                graph_uuid, group_uuid = args['graph_uuid'], args['group_uuid']
                self.backend.group_get_result(graph_uuid=graph_uuid, group_uuid=group_uuid)
            elif command == commands.BACKEND_CHAIN_GET_RESULT:
                graph_uuid, chain_uuid = args['graph_uuid'], args['chain_uuid']
                self.backend.chain_get_result(graph_uuid=graph_uuid, chain_uuid=chain_uuid)

    def start(self):
        super().start()

    def result_get(self, task_uuid, graph_uuid=None):
        self.cmd.put(commands.backend_result_get(task_uuid=task_uuid, graph_uuid=graph_uuid))
        return self.cmd_out.get()

    def result_set(self, task_uuid, result, graph_uuid=None):
        self.cmd.put(commands.backend_result_set(task_uuid=task_uuid, result=result, graph_uuid=graph_uuid))
        self.cmd_out.get()

    def result_del(self, task_uuid):
        self.cmd.put(commands.backend_result_del(task_uuid=task_uuid))

    def results_del(self, *task_uuids):
        self.cmd.put(commands.backend_results_del(*task_uuids))

    def graph_init(self, graph_dict):
        self.cmd.put(commands.backend_graph_init(graph_dict=graph_dict))
        self.cmd_out.get()

    def graph_get(self, graph_uuid):
        self.cmd.put(commands.backend_graph_get(graph_uuid=graph_uuid))
        return self.cmd_out.get()

    def graph_update(self, graph_uuid, verticies, graph_state=None, error_task_uuid=None, short_error=None):
        self.cmd.put(commands.backend_graph_update(graph_uuid, verticies, graph_state, error_task_uuid, short_error))
        return self.cmd_out.get()

    def graph_result_set(self, graph_uuid, task_uuid, result):
        self.cmd.put(commands.backend_graph_result_set(graph_uuid=graph_uuid, task_uuid=task_uuid, result=result))
        return self.cmd_out.get()

    def is_group_completed(self, graph_uuid, group_uuid, expected):
        self.cmd.put(
            commands.backend_is_group_completed(graph_uuid=graph_uuid, group_uuid=group_uuid, expected=expected))
        return self.cmd_out.get()

    def group_add_result(self, graph_uuid, group_uuid, task_uuid, result):
        self.cmd.put(
            commands.backend_group_add_result(graph_uuid=graph_uuid, group_uuid=group_uuid, task_uuid=task_uuid,
                                              result=result))
        return self.cmd_out.get()

    def group_get_result(self, graph_uuid, group_uuid):
        self.cmd.put(commands.backend_group_get_result(graph_uuid=graph_uuid, group_uuid=group_uuid))
        return self.cmd_out.get()

    def chain_get_result(self, graph_uuid, chain_uuid):
        self.cmd.put(commands.backend_chain_get_result(graph_uuid=graph_uuid, chain_uuid=chain_uuid))
        return self.cmd_out.get()



import logging
import signal
import threading
import time

import pika
from pika import exceptions

import gromozeka.app
from gromozeka.brokers.base import BrokerInterface
from gromozeka.primitives import Task

APP_ID = 'gromozeka'
CONSUMER_ID_FORMAT = "{exchange}.{exchange_type}.{queue}.{routing_key}"
PIKA_CONNECTION_EXCEPTIONS = (exceptions.AMQPConnectionError,
                              exceptions.ConnectionClosed)
PIKA_CHANNEL_EXCEPTIONS = (
    exceptions.AMQPConnectionError,
    exceptions.ConnectionClosed,
    exceptions.ChannelClosed,
    OverflowError,
    AttributeError)


def format_consumer_id_from_broker_point(broker_point):
    return CONSUMER_ID_FORMAT.format(exchange=broker_point.exchange, exchange_type=broker_point.exchange_type,
                                     queue=broker_point.queue, routing_key=broker_point.routing_key)


class RabbitMQPikaAdaptee(BrokerInterface):
    """Pika adapted broker

    Attributes:
        reconnect_max_retries(int): Maximum attempts to reconnect
        reconnect_retry_countdown(int|float): Maximum attempts to reconnect
        connection(pika.SelectConnection): Pika connection

    """

    def __init__(self, app):
        """

        Args:
            app (gromozeka.app.app.Gromozeka):
        """
        self.reconnect_max_retries = None
        self.reconnect_retry_countdown = None
        self.connection = None
        self._consumers = {}
        self._url = None
        self._closing = False
        self.reconnecting = False
        self._prefetch_lock = threading.Lock()
        super().__init__(app=app)

    @staticmethod
    def worker_run(self):
        """

        Args:
            self(gromozeka.brokers.base.BrokerAdapter):

        Returns:

        """
        self.broker.connection = self.broker.connect()
        self.broker.connection.ioloop.start()

    def stop(self):
        self.stop_serve()

    def configure(self):
        self._url = self.app.config.broker_url
        self.reconnect_max_retries = self.app.config.broker_reconnect_max_retries
        self.reconnect_retry_countdown = self.app.config.broker_reconnect_retry_countdown

    def get_consumer(self, broker_point):
        return self._consumers[format_consumer_id_from_broker_point(broker_point)]

    def task_register(self, broker_point, task_id, options, deserializator):
        consumer_id = format_consumer_id_from_broker_point(broker_point)
        try:
            consumer = self._consumers[consumer_id]
        except KeyError:
            consumer = Consumer(
                exchange=broker_point.exchange,
                exchange_type=broker_point.exchange_type,
                queue=broker_point.queue,
                routing_key=broker_point.routing_key,
                deserializator=deserializator)
            self._consumers[consumer_id] = consumer
        consumer.add_task(task_id)

    def task_send(self, task_uuid, request, broker_point, reply_to=None):
        consumer = self._consumers[format_consumer_id_from_broker_point(broker_point)]
        if reply_to:
            exchange, routing_key = reply_to.exchange, reply_to.routing_key
        else:
            exchange, routing_key = None, None
        consumer.publish(task_uuid=task_uuid, request=request, exchange=exchange, routing_key=routing_key)

    def task_send_delayed(self, task_uuid, request, broker_point, delay):
        publisher = DelayedMessagePublisher(task_uuid=task_uuid, broker_connection=self.connection, request=request,
                                            original_exchange=broker_point.exchange,
                                            original_routing_key=broker_point.routing_key, delay=delay)
        publisher.publish()

    def task_done(self, task_uuid, broker_point, delivery_tag):
        consumer = self._consumers[format_consumer_id_from_broker_point(broker_point)]
        consumer.acknowledge_message(task_uuid, delivery_tag)

    def task_reject(self, task_uuid, broker_point, delivery_tag):
        consumer = self._consumers[format_consumer_id_from_broker_point(broker_point)]
        consumer.reject_message(task_uuid, delivery_tag)

    def on_pool_size_changed(self):
        with self._prefetch_lock:
            for cname, consumer in self._consumers.items():
                new_prefetch_count = 0
                for _, task in self.app.registry.items():
                    consumer_id = format_consumer_id_from_broker_point(task.broker_point)
                    if consumer_id == cname:
                        new_prefetch_count += task.pool.size
                if consumer.prefetch_count != new_prefetch_count:
                    consumer.change_prefetch_count(new_prefetch_count)
                    if consumer.prefetch_count == 0:
                        consumer.stop_consuming()

    def wait_for_start(self):
        wfs_consumers = [c for _, c in self._consumers.items()]
        while wfs_consumers:
            for n, c in enumerate(wfs_consumers):
                if c.is_consuming:
                    del (wfs_consumers[n])

    def connect(self):
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.

        Returns:
            pika.SelectConnection:

        """
        self.logger.info('connecting to %s', self._url)
        return pika.adapters.SelectConnection(parameters=pika.URLParameters(self._url),
                                              on_open_callback=self.on_connection_open,
                                              stop_ioloop_on_close=False)

    def on_connection_open(self, unused_connection):
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.

        Args:
            unused_connection(pika.SelectConnection): Unused connection

        """
        self.logger.info('connection opened')
        self.add_on_connection_close_callback()
        for _, consumer in self._consumers.items():
            consumer.connection = self.connection
            consumer.open_channel()

    def add_on_connection_close_callback(self):
        """This method adds an on close callback that will be invoked by pika
        when RabbitMQ closes the connection to the publisher unexpectedly.

        """
        self.logger.debug('adding connection close callback')
        self.connection.add_on_close_callback(self.on_connection_closed)

    def on_connection_closed(self, connection, reply_code, reply_text):
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.

        Args:
            connection(pika.connection.Connection): The closed connection obj
            reply_code(int): Reply code if given
            reply_text(str): The server provided reply_text if given

        """
        for _, consumer in self._consumers.items():
            consumer._channel = None
        if self._closing:
            self.connection.ioloop.stop()
            self.logger.info('connection closed')
        else:
            if self.reconnect_max_retries:
                self.logger.warning('starting reconnection process in %d seconds: (%s) %s',
                                    self.reconnect_retry_countdown, reply_code, reply_text)
                self.connection.add_timeout(self.reconnect_retry_countdown, self.reconnect)
                return
            self.logger.fatal("broker connection problems: (%s) %s", reply_code, reply_text)
            self.stop()
            self.app.stop_signal(signal.SIGTERM)

    def reconnect(self):
        """Will be invoked by the IOLoop timer if the connection is
        closed. See the on_connection_closed method.

        """
        # This is the old connection IOLoop instance, stop its ioloop
        self.connection.ioloop.stop()
        if self._closing or self.connection.is_open:
            return
        attempts = self.reconnect_max_retries
        stop_exc = None
        while not self.app.is_closing:
            self.reconnecting = True
            # Create a new connection
            try:
                self.connection = self.connect()
                # There is now a new connection, needs a new ioloop to run
                self.connection.ioloop.start()
                self.reconnecting = False
                return
            except PIKA_CONNECTION_EXCEPTIONS as e:
                self.logger.warning(e)
                attempts -= 1
                if attempts <= 0:
                    stop_exc = "broker connection problems, max reconnects exceeded: %s" % e
                    break
                time.sleep(self.reconnect_retry_countdown)
                continue
            except Exception as e:
                stop_exc = "unhandled exception while reconnecting: %s" % e
                break
        if self.app.is_closing:
            return
        self.logger.fatal(stop_exc)
        self.stop()
        self.app.stop_signal(signal.SIGTERM)

    def stop_serve(self):
        """Cleanly shutdown the connection to RabbitMQ by stopping the consumers
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.

        """
        self._closing = True
        wfs_consumers = []
        for _, consumer in self._consumers.items():
            if not consumer.channel or consumer.channel.is_closed or consumer.channel.is_closing:
                continue
            consumer.stop_consuming()
            wfs_consumers.append(consumer)
        while wfs_consumers:
            for n, c in enumerate(wfs_consumers):
                if c.channel.is_closed:
                    del (wfs_consumers[n])
        if self.connection.is_open:
            self.close_connection()
        self.connection.ioloop.stop()

    def close_connection(self):
        """This method closes the connection to RabbitMQ."""
        self.logger.info('closing connection')
        self.connection.close()


class DelayedMessagePublisher:
    exchange = 'delayed'
    exchange_type = 'direct'

    def __init__(self, task_uuid, broker_connection, request, original_exchange, original_routing_key, delay):
        self.logger = logging.getLogger('gromozeka.broker.consumer.delayed_publisher')
        self.task_uuid = task_uuid
        self._connection = broker_connection
        self.ttl = delay * 1000
        self.queue = 'delay.%d.%s.%s' % (self.ttl, original_exchange, original_routing_key)
        self.routing_key = self.queue
        self.original_exchange = original_exchange
        self.original_routing_key = original_routing_key
        self.request = request
        self.channel = None

    def open_channel(self):
        """Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.

        """
        self.logger.debug('creating a new channel')
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.

        Since the channel is now open, we'll declare the exchange to use.

        Args:
            channel(pika.channel.Channel): The closed channel

        """
        self.logger.debug('channel opened')
        self.channel = channel
        self.add_on_channel_close_callback()
        self.setup_exchange(self.exchange)

    def add_on_channel_close_callback(self):
        """This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.

        """
        self.logger.debug('adding channel close callback')
        self.channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reply_code, reply_text):
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.

        Args:
            channel(pika.channel.Channel): The closed channel
            reply_code(int): The numeric reason the channel was closed
            reply_text(str): The text reason the channel was closed
        """
        self.logger.debug('channel %i was closed: (%s) %s', channel, reply_code, reply_text)

    def setup_exchange(self, exchange_name):
        """Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
        command. When it is complete, the on_exchange_declareok method will
        be invoked by pika.

        Args:
            exchange_name(str): The name of the exchange to declare

        """
        self.logger.debug('declaring exchange %s', exchange_name)
        self.channel.exchange_declare(self.on_exchange_declareok, exchange_name, self.exchange_type)

    def on_exchange_declareok(self, unused_frame):
        """Invoked by pika when RabbitMQ has finished the Exchange.Declare RPC
        command.

        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame

        """
        self.logger.debug('exchange declared')
        self.setup_queue(self.queue)

    def setup_queue(self, queue_name, ):
        """Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
        command. When it is complete, the on_queue_declareok method will
        be invoked by pika.

        :param str|unicode queue_name: The name of the queue to declare.

        """
        self.logger.debug('declaring queue %s', queue_name)
        self.channel.queue_declare(self.on_queue_declareok, queue=queue_name, durable=True, arguments={
            'x-dead-letter-exchange': self.original_exchange,  # Exchange used to transfer the message from A to B.
            'x-dead-letter-routing-key': self.original_routing_key,  # Name of the queue message transferred to.
            'x-message-ttl': self.ttl,  # Delay until the message is transferred in milliseconds.
            'x-expires': self.ttl * 2
        })

    def on_queue_declareok(self, method_frame):
        """Method invoked by pika when the Queue.Declare RPC call made in
        setup_queue has completed. In this method we will bind the queue
        and exchange together with the routing key by issuing the Queue.Bind
        RPC command. When this command is complete, the on_bindok method will
        be invoked by pika.

        Args:
            method_frame(pika.frame.Method): The Queue.DeclareOk frame

        """
        self.logger.debug('binding %s to %s with %s', self.exchange, self.queue, self.routing_key)
        self.channel.queue_bind(self.on_bindok, self.queue, self.exchange, self.routing_key)

    def on_bindok(self, unused_frame):
        """Invoked by pika when the Queue.Bind method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.

        Args:
            unused_frame(pika.frame.Method): The Queue.BindOk response frame

        """
        self.logger.debug('queue bound')
        try:
            self.channel.basic_publish(self.exchange, self.queue, self.request,
                                       properties=pika.BasicProperties(delivery_mode=2, app_id=APP_ID))
        except PIKA_CHANNEL_EXCEPTIONS:
            self.logger.warning('connection lost on message retry, retry will be ignored')
            return
        self.close_channel()

    def add_on_cancel_callback(self):
        """Add a callback that will be invoked if RabbitMQ cancels the consumer
        for some reason. If RabbitMQ does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.

        """
        self.logger.debug('adding consumer cancellation callback')
        self.channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.

        Args:
            method_frame(pika.frame.Method): The Basic.Cancel frame

        """
        self.logger.info('consumer was cancelled remotely, shutting down: %r', method_frame)
        if self.channel:
            self.channel.close()

    def on_cancelok(self, unused_frame):
        """This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.

        Args:
            unused_frame(pika.frame.Method): The Basic.CancelOk frame

        """
        self.logger.debug('RabbitMQ acknowledged the cancellation of the consumer')
        self.close_channel()

    def close_channel(self):
        """Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.

        """
        self.logger.debug('closing the channel')
        self.channel.close()

    def publish(self):
        self.open_channel()


class Consumer:
    def __init__(self, exchange, exchange_type, queue, routing_key, deserializator=None):
        consumer_id = "%s.%s.%s.%s" % (exchange, exchange_type, queue, routing_key)
        self.logger = logging.getLogger(
            '{}.broker.consumer.{}'.format(gromozeka.app.get_app().config.app_id, consumer_id))
        self.consumer_tag = consumer_id
        self.exchange = exchange
        self.exchange_type = exchange_type
        self.queue = queue
        self.routing_key = routing_key
        self.tasks = {}
        self.channel = None
        self.connection = None
        self._closing = False
        self.is_consuming = False
        self.prefetch_count = 0
        self.stop_lock = threading.Lock()
        self.deserializator = deserializator
        from gromozeka.brokers import LatexSema
        self.sema = LatexSema(threading.Semaphore, self.prefetch_count)

    def add_task(self, task_id):
        self.tasks[task_id] = gromozeka.app.get_app().get_task(task_id)

    def open_channel(self):
        """Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.

        """
        self.logger.debug('creating a new channel')
        self.connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.

        Since the channel is now open, we'll declare the exchange to use.

        Args:
            channel(pika.channel.Channel): The closed channel

        """
        self.logger.debug('channel opened')
        self.channel = channel
        self.add_on_channel_close_callback()
        self.setup_exchange(self.exchange)

    def add_on_channel_close_callback(self):
        """This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.

        """
        self.logger.debug('adding channel close callback')
        self.channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reply_code, reply_text):
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.

        Args:
            channel(pika.channel.Channel): The closed channel
            reply_code(int): The numeric reason the channel was closed
            reply_text(str): The text reason the channel was closed
        """
        self.logger.debug('channel %i was closed: (%s) %s', channel, reply_code, reply_text)

    def setup_exchange(self, exchange_name):
        """Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
        command. When it is complete, the on_exchange_declareok method will
        be invoked by pika.

        Args:
            exchange_name(str): The name of the exchange to declare

        """
        self.logger.debug('declaring exchange %s', exchange_name)
        self.channel.exchange_declare(self.on_exchange_declareok, exchange_name, self.exchange_type)

    def on_exchange_declareok(self, unused_frame):
        """Invoked by pika when RabbitMQ has finished the Exchange.Declare RPC
        command.

        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame

        """
        self.logger.debug('exchange declared')
        self.setup_queue(self.queue)

    def setup_queue(self, queue_name):
        """Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
        command. When it is complete, the on_queue_declareok method will
        be invoked by pika.

        :param str|unicode queue_name: The name of the queue to declare.

        """
        self.logger.debug('declaring queue %s', queue_name)
        self.channel.queue_declare(self.on_queue_declareok, queue_name, durable=True)

    def on_queue_declareok(self, method_frame):
        """Method invoked by pika when the Queue.Declare RPC call made in
        setup_queue has completed. In this method we will bind the queue
        and exchange together with the routing key by issuing the Queue.Bind
        RPC command. When this command is complete, the on_bindok method will
        be invoked by pika.

        Args:
            method_frame(pika.frame.Method): The Queue.DeclareOk frame

        """
        self.logger.debug('binding %s to %s with %s', self.exchange, self.queue, self.routing_key)
        self.channel.queue_bind(self.on_bindok, self.queue, self.exchange, self.routing_key)

    def on_bindok(self, unused_frame):
        """Invoked by pika when the Queue.Bind method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.

        Args:
            unused_frame(pika.frame.Method): The Queue.BindOk response frame

        """
        self.logger.debug('queue bound')
        self.start_consuming()

    def start_consuming(self):
        """This method sets up the consumer by first calling
        add_on_cancel_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming. The on_message method is passed in as a callback pika
        will invoke when a message is fully received.

        """
        self.logger.debug('issuing consumer related RPC commands')
        self.channel.basic_qos(prefetch_count=self.prefetch_count)
        self.add_on_cancel_callback()
        self.consumer_tag = self.channel.basic_consume(self.on_message, self.queue)
        self.is_consuming = True

    def add_on_cancel_callback(self):
        """Add a callback that will be invoked if RabbitMQ cancels the consumer
        for some reason. If RabbitMQ does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.

        """
        self.logger.debug('adding consumer cancellation callback')
        self.channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.

        Args:
            method_frame(pika.frame.Method): The Basic.Cancel frame

        """
        self.logger.info('consumer was cancelled remotely, shutting down: %r', method_frame)
        if self.channel:
            self.channel.close()

    def change_prefetch_count(self, new_size):
        if self.channel and self.channel.is_open:
            self.channel.basic_qos(prefetch_size=new_size if self.prefetch_count < new_size else 0,
                                   prefetch_count=new_size)
        self.logger.info('prefetch count changed from %s to %s', self.prefetch_count, new_size)
        self.prefetch_count = new_size
        self.sema.change(new_size)

    def publish(self, task_uuid, request, exchange=None, routing_key=None):
        """

        Args:
            task_uuid:
            request:
            exchange(str):
            routing_key(str):

        Returns:

        """
        if exchange and routing_key:
            exchange, routing_key = exchange, routing_key
        else:
            exchange, routing_key = self.exchange, self.routing_key
        try:
            self.channel.basic_publish(exchange=exchange, routing_key=routing_key, body=request,
                                       properties=pika.BasicProperties(delivery_mode=2, app_id=APP_ID))
        except PIKA_CHANNEL_EXCEPTIONS:
            self.logger.warning('connection lost on message publish')

    def on_message(self, unused_channel, basic_deliver, properties, body):
        """Invoked by pika when a message is delivered from RabbitMQ. The
        channel is passed for your convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.

        Args:
            unused_channel(pika.channel.Channel): The channel object
            basic_deliver(pika.Spec.Basic.Deliver): Basic deliver method
            properties(pika.Spec.BasicProperties): Properties
            body(str|bytes): The message body

        """
        self.sema.acquire()
        task = Task.from_proto(body, deserializator=self.deserializator if properties.app_id != APP_ID else None)

        self.logger.info('received message #%s, task_uuid <%s> from %s', basic_deliver.delivery_tag, task.uuid,
                         properties.app_id)
        task.delivery_tag = basic_deliver.delivery_tag
        task.on_receive()
        self.tasks[task.task_id].pool.worker_queue.put(task.request.SerializeToString())

    def acknowledge_message(self, task_uuid, delivery_tag):
        """Acknowledge the message delivery from RabbitMQ by sending a
        Basic.Ack RPC method for the delivery tag.

        Args:
            task_uuid:
            delivery_tag(int): The delivery tag from the Basic.Deliver frame

        """
        try:
            self.channel.basic_ack(delivery_tag)
        except PIKA_CHANNEL_EXCEPTIONS:
            self.logger.warning('connection lost on message acknowledge #%d' % delivery_tag)
        finally:
            self.sema.release()
        self.logger.info('acknowledged message #%d, task_uuid <%s>', delivery_tag, task_uuid)

    def reject_message(self, task_uuid, delivery_tag, requeue=False):
        """Reject message by it`s delivery tag

        Args:
            task_uuid(str):
            delivery_tag(int):
            requeue(bool): If `True` message will reject with requeue
        """
        self.logger.info('rejecting message %s, task_uuid <%s> from gromozeka', delivery_tag, task_uuid)
        try:
            self.channel.basic_reject(delivery_tag, requeue)
        except PIKA_CHANNEL_EXCEPTIONS:
            self.logger.warning('connection lost on message reject #%d' % delivery_tag)
        finally:
            self.sema.release()

    def stop_consuming(self):
        """Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.

        """
        self.stop_lock.acquire()
        if not self.channel:
            return
        if self.channel.is_closing or self.channel.is_closed:
            return
        self.logger.debug('sending a Basic.Cancel RPC command to RabbitMQ')
        self.channel.basic_cancel(self.on_cancelok, self.consumer_tag)

    def on_cancelok(self, unused_frame):
        """This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.

        Args:
            unused_frame(pika.frame.Method): The Basic.CancelOk frame

        """
        self.logger.debug('broker acknowledged the cancellation of the consumer')
        self.close_channel()
        self.stop_lock.release()

    def close_channel(self):
        """Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.

        """
        self.logger.debug('closing the channel')
        self.channel.close()



import logging
import multiprocessing
import queue
import threading

import gromozeka.app
from gromozeka.concurrency import commands
from gromozeka.concurrency.worker import ThreadWorker, ProcessWorker


class Pool(threading.Thread):
    """Pool of workers

    Attributes:
        logger(:obj:`logging.Logger`): Class logger
        worker_queue(:obj:`multiprocessing.Queue`): Queue to receive input requests
        cmd(:obj:`multiprocessing.Queue`): Command queue
        cmd_out(:obj:`multiprocessing.Queue`): Command result queue
        max_workers(:obj:`int`): Number of workers to start
        worker_class(:class:`gromozeka.concurrency.Worker`): worker class
        workers(:obj:`list` of :obj:`gromozeka.concurrency.Worker`): list of worker instances
        _stop_event(:obj:`multiprocessing.Event`): Stop event

    Args:
        max_workers(:obj:`int`, optional): Number of workers to start
        worker_class(:class:`gromozeka.concurrency.Worker`, optional): Worker class
    """

    __slots__ = ['logger', 'init_max_workers', 'worker_cls', 'worker_queue', 'cmd', 'cmd_out', '_stop_event', 'workers',
                 'is_on_stop']

    def __init__(self, max_workers=1, worker_class=ThreadWorker, logger=None):
        self.logger = logger or logging.getLogger("%s.pool" % gromozeka.app.get_app().config.app_id)
        self.init_max_workers = max_workers
        self.worker_cls = worker_class
        self.worker_queue = multiprocessing.Queue() if issubclass(self.worker_cls, ProcessWorker) else queue.Queue()
        from gromozeka.brokers import BrokerAdapter
        from gromozeka.backends import BackendAdapter
        self.cmd = multiprocessing.Queue() if isinstance(self, (BrokerAdapter, BackendAdapter)) else queue.Queue()
        self.cmd_out = multiprocessing.Queue() if isinstance(self, (BrokerAdapter, BackendAdapter)) else queue.Queue()
        self._stop_event = threading.Event()
        self.is_on_stop = False
        self.workers = {}
        self.lock = threading.Lock()
        super().__init__()

    def listen_cmd(self):
        """Listen to commands

        """
        while not self._stop_event.is_set():
            try:
                command, args = self.cmd.get(True, 0.05)
            except queue.Empty:
                continue
            if command == commands.POOL_SHRINK:
                self._shrink(args)
            elif command == commands.POOL_GROW:
                self._grow(args)
            elif command == commands.POOL_REMOVE_WORKER:
                self._remove_worker(args)
            elif command == commands.POOL_STOP:
                self.stop_()

    def start(self):
        """Start pool

        """
        with self.lock:
            super().start()
            self.cmd_out.get()
        gromozeka.app.get_app().broker_adapter.on_pool_size_changed()

    def run(self):
        """Start :attr:`workers` in threads or processes

        """
        self.logger.info("start")
        self._grow(self.init_max_workers)
        self.listen_cmd()

    def stop(self):
        """Stop pool and child :attr:`workers`

        """
        with self.lock:
            self.cmd.put(commands.pool_stop())
            self.cmd_out.get()
        self.logger.info("stop")

    def stop_(self):
        self.is_on_stop = True
        self._shrink(len(self.workers))
        self._stop_event.set()

    def grow(self, n):
        """Grow pool by `n` number of workers

        Args:
            n(int): Number to grow
        """
        with self.lock:
            self.cmd.put(commands.pool_grow(n))
            self.cmd_out.get()
        gromozeka.app.get_app().broker_adapter.on_pool_size_changed()

    def _grow(self, n):
        workers = []
        for _ in range(n):
            worker = self.worker_cls(pool=self)
            worker.start()
            self.workers[worker.ident] = worker
            workers.append(worker)
        while workers:
            for worker in workers:
                if worker.is_alive():
                    workers.remove(worker)
        self.cmd_out.put(True)

    def shrink(self, n):
        """Shrink pool by `n` number of workers

        Args:
            n(int): Number to grow
        """
        with self.lock:
            self.cmd.put(commands.pool_shrink(n))
            self.cmd_out.get()
        gromozeka.app.get_app().broker_adapter.on_pool_size_changed()

    def _shrink(self, n):
        workers = []
        for ident in list(self.workers.keys())[0:n]:
            worker = self.workers.pop(ident)
            worker.stop()
            workers.append(worker)
        while workers:
            for worker in workers:
                if not worker.is_alive():
                    workers.remove(worker)
        self.cmd_out.put(True)

    def remove_worker(self, worker_ident):
        """Remove worker from pool by it`s ident

        Args:
            worker_ident: Worker identification
        """
        # with self.lock:
        self.cmd.put(commands.pool_remove_worker(worker_ident))
        # self.cmd_out.get()
        gromozeka.app.get_app().broker_adapter.on_pool_size_changed()

    def _remove_worker(self, worker_ident):
        del self.workers[worker_ident]
        self.cmd_out.put(True)

    @property
    def size(self):
        """Not command do not use this in workers

        Returns:

        """
        with self.lock:
            return len(self.workers)



import inspect
import json
from collections import deque
from uuid import uuid4

import gromozeka
import gromozeka.primitives.base
from gromozeka.primitives.task_pb2 import Task as ProtoTask, ReplyToBrokerPoint as ProtoReplyTo

GROUP = 'GROUP'  # Group operation for parallel computation
CHAIN = 'CHAIN'  # Chain operation for execute tasks one by one
ACHAIN = 'ACHAIN'  # Argumented chain operation for execute tasks one by one with arguments from previous task


class Graph:
    first_uuid = None
    last_uuid = None

    def __init__(self, *args):
        """

        Args:
            args(:obj:`list of :obj:`Vertex`): list of vertexes to make Graph

        """
        self.verticies = {}
        self.uuid = None
        self.state = gromozeka.primitives.base.RECEIVED
        self.error_task_uuid = None
        self.short_error = None
        if args:
            self.first_uuid = args[0].uuid
            self.last_uuid = args[-1].uuid
        for vertex in args:
            self.verticies[vertex.uuid] = vertex

    @property
    def last(self):
        """

        Returns:
            :obj:`Vertex`: Last :obj:`Vertex` of :obj:`Graph`
        """
        return self.vertex_by_uuid(self.last_uuid)

    @property
    def first(self):
        """

        Returns:
            :obj:`Vertex`: First :obj:`Vertex` of :obj:`Graph`

        """
        return self.verticies[self.first_uuid]

    def __or__(self, other):
        """

        Args:
            other(Vertex or Graph):

        Returns:
            Graph:
        """
        return self._op(other, GROUP)

    def __gt__(self, other):
        """

        Args:
            other(Vertex or Graph):

        Returns:
            Graph:
        """
        return self._op(other, CHAIN)

    def __rshift__(self, other):
        return self._op(other, ACHAIN)

    def _op(self, other, op_name):
        """

        Args:
            other(Vertex or Graph): other graph or vertex
            op_name(str): operation name

        Returns:
            Graph:
        """
        if isinstance(other, Graph):
            if op_name == GROUP and self.last.operation == GROUP:
                operation_vertex = self.last
            else:
                operation_vertex = Vertex(task_id=op_name, op_name=op_name)
                self.last.add_edge(operation_vertex)
            other.last.add_edge(operation_vertex)
            self.verticies[operation_vertex.uuid] = operation_vertex
            self.verticies.update(other.verticies)
            if operation_vertex.operation == ACHAIN:
                [setattr(self.vertex_by_uuid(child), 'args_from', self.last.uuid) for child in other.last.children]
            self.last_uuid = self.uuid = operation_vertex.uuid
        else:
            if op_name == CHAIN:
                self._chain(other)
            elif op_name == ACHAIN:
                self._achain(other)
            elif op_name == GROUP:
                self._group(other)
        return self

    def _chain(self, other):
        """

        Args:
            other(Vertex or Graph): other graph or vertex
        """
        self.verticies[other.uuid] = other
        operation_vertex = Vertex(task_id=CHAIN, op_name=CHAIN)
        self.last.add_edge(operation_vertex)
        other.add_edge(operation_vertex)
        self.verticies[operation_vertex.uuid] = operation_vertex
        self.last_uuid = self.uuid = operation_vertex.uuid

    def _achain(self, other):
        """

        Args:
            other(Vertex or Graph): other graph or vertex
        """
        self.verticies[other.uuid] = other
        operation_vertex = Vertex(task_id=ACHAIN, op_name=ACHAIN)
        self.last.add_edge(operation_vertex)
        other.add_edge(operation_vertex)
        other.args_from = self.last.uuid
        self.verticies[operation_vertex.uuid] = operation_vertex
        self.last_uuid = self.uuid = operation_vertex.uuid

    def _group(self, other):
        """

        Args:
            other(Vertex or Graph): other graph or vertex
        """
        self.verticies[other.uuid] = other
        if self.last.operation == GROUP:
            other.add_edge(self.last)
        else:
            operation_vertex = Vertex(task_id=GROUP, op_name=GROUP)
            self.last.add_edge(operation_vertex)
            other.add_edge(operation_vertex)
            self.verticies[operation_vertex.uuid] = operation_vertex
            self.last_uuid = self.uuid = operation_vertex.uuid

    def vertex_by_uuid(self, uuid):
        """Get :obj:`Vertex` by uuid

        Args:
            uuid(str):

        Returns:
            Vertex:
        """
        return self.verticies[uuid]

    def is_group_part(self, vertex):
        if not vertex.parent:
            return False
        return self.vertex_by_uuid(vertex.parent).operation == GROUP

    def is_chain_part(self, vertex):
        if not vertex.parent:
            return False
        return self.vertex_by_uuid(vertex.parent).operation in (CHAIN, ACHAIN)

    def is_chain_tail(self, vertex):
        if not self.is_chain_part(vertex):
            return False
        return vertex.weight == 0

    def as_dict(self):
        d = {}
        vertex_members = dir(Vertex())
        for gk, gv in self.__dict__.items():
            if gk == 'verticies':
                verticies = {}
                for vk, vv in gv.items():
                    vertex = {}
                    for k in dir(vv):
                        v = getattr(vv, k)
                        if k not in vertex_members or k.startswith('__') or inspect.isroutine(v):
                            continue
                        vertex[k] = v
                    verticies.update({vk: vertex})
                d[gk] = verticies
                continue
            d[gk] = gv
        return d

    @classmethod
    def from_dict(cls, data):
        graph = cls()
        verticies = {}
        for gk, gv in data.items():
            if hasattr(graph, gk):
                setattr(graph, gk, gv)
                continue
            vertex = Vertex()
            for vk, vv in json.loads(gv).items():
                setattr(vertex, vk, vv)
            verticies.update({gk: vertex})
        setattr(graph, 'verticies', verticies)
        return graph

    def apply_async(self, graph_uuid=None):
        app = gromozeka.get_app()
        if graph_uuid:
            self.uuid = graph_uuid
        app.backend_adapter.graph_init(self.as_dict())
        for v in self.next_vertex():
            v.apply(self.uuid)

    def next_vertex(self, from_uuid=None):
        stack = deque([self.vertex_by_uuid(from_uuid) if from_uuid else self.last])
        while stack:
            cur = stack[-1]
            if cur.operation:
                if cur.state == gromozeka.primitives.base.SUCCESS:
                    if cur.parent:
                        stack.appendleft(self.vertex_by_uuid(cur.parent))
                else:
                    children = [self.vertex_by_uuid(ch) for ch in cur.children]
                    if cur.operation == GROUP:
                        for ch in children:
                            if ch.state == gromozeka.primitives.base.SUCCESS:
                                break
                        else:
                            stack.extendleft(children)
                    else:
                        for ch in children:
                            if ch.state != gromozeka.primitives.base.SUCCESS:
                                stack.appendleft(ch)
                                break
            elif cur.state != gromozeka.primitives.base.SUCCESS:
                yield cur
            else:
                if cur.parent:
                    stack.appendleft(self.vertex_by_uuid(cur.parent))
            stack.pop()

    def draw(self):
        """Draw workflow graph

        """

        import networkx as nx
        import matplotlib as mpl
        mpl.use('Qt5Agg')
        from matplotlib import pyplot as plt
        nx_graph = nx.DiGraph()
        node_labels = {}
        for uuid, v in self.verticies.items():
            if v.parent:
                nx_graph.add_edge(v.uuid, v.to, weight=v.weight)
            node_labels[v.uuid] = '{}{}'.format(v.task_id or v.operation,
                                                '(R)' if (v.operation and not v.parent) else '')
            nx_graph.add_node(v.uuid, label=v.task_id or v.operation)
        pos = nx.nx_pydot.graphviz_layout(nx_graph)
        nx.draw(nx_graph, pos=pos, node_size=3000, node_color='lightgreen', labels=node_labels)
        edge_labels = nx.get_edge_attributes(nx_graph, 'weight')
        nx.draw_networkx_edge_labels(nx_graph, pos, edge_labels=edge_labels)
        plt.show()


class Vertex:
    def __init__(self, uuid=None, task_id=None, op_name=None, args=None, kwargs=None, bind=None, reply_to_exchange=None,
                 reply_to_routing_key=None):
        self.uuid = uuid or str(uuid4())
        self.bind = bind or False
        self.parent = None
        self.children = []
        self.state = gromozeka.primitives.base.PENDING
        self.operation = op_name
        self.fr = None
        self.to = None
        self.weight = None
        self.task_id = task_id
        self.args = args
        self.kwargs = kwargs
        self.error = None
        self.exc = None
        self.args_from = None
        self.reply_to_exchange = reply_to_exchange
        self.reply_to_routing_key = reply_to_routing_key

    def as_dict(self):
        d = {}
        for vk, vv in self.__dict__.items():
            d[vk] = vv
        return d

    def apply(self, graph_uuid):
        app = gromozeka.get_app()
        bp = app.get_task(self.task_id).broker_point
        if self.reply_to_exchange and self.reply_to_routing_key:
            reply_to = ProtoReplyTo(exchange=self.reply_to_exchange,
                                    routing_key=self.reply_to_routing_key)
        else:
            reply_to = None
        proto_task = ProtoTask(uuid=self.uuid, task_id=self.task_id, graph_uuid=graph_uuid, reply_to=reply_to)
        if self.args:
            proto_task.args = json.dumps(self.args)
        if self.kwargs:
            proto_task.kwargs = json.dumps(self.kwargs)

        app.broker_adapter.task_send(task_uuid=self.uuid, request=proto_task.SerializeToString(), broker_point=bp)

    def add_edge(self, to):
        """

        Args:
            to(Vertex):
        """
        weight = 0 if (to.children and to.operation in (CHAIN, ACHAIN)) or to.operation == GROUP else 1
        to.children.append(self.uuid)
        self.fr = self.uuid
        self.to = to.uuid
        self.weight = weight
        self.parent = to.uuid

    def __or__(self, other):
        """

        Args:
            other(Vertex or Graph):

        Returns:
            Graph:
        """
        return Graph(self).__or__(other)

    def __gt__(self, other):
        """

        Args:
            other(Vertex or Graph):

        Returns:
            Graph:
        """
        return Graph(self).__gt__(other)

    def __rshift__(self, other):
        """

        Args:
            other(Vertex or Graph)::

        Returns:
            Graph:
        """
        return Graph(self).__rshift__(other)



import functools
import logging
import multiprocessing
import queue
import threading
import traceback

import gromozeka.app
from gromozeka.exceptions import Retry, MaxRetriesExceedException
from gromozeka.primitives import Task


class Worker:
    """Worker class for `gromozeka.concurrency.Pool`. Uses as mixin.

    Attributes:
        logger(logging.Logger): class logger
        pool(gromozeka.concurrency.Pool): parent pool instance
        _stop_event(multiprocessing.Event or threading.Event)

    Args:
        pool(gromozeka.concurrency.Pool): Parent pool instance
    """

    name = None  # multiprocessing or threading part
    ident = None  # multiprocessing or threading part

    def __init__(self, pool):
        self.logger = logging.getLogger("%s.pool.worker" % gromozeka.app.get_app().config.app_id)
        self.pool = pool
        self._stop_event = None

    def run(self):
        """Run worker in thread or process

        Raises:
            exceptions.MaxRetriesExceedException:
            BaseException:
        """
        self.logger.info("start")
        exc = None
        while not self._stop_event.is_set():
            try:
                # get request object from queue
                task_proto = self.pool.worker_queue.get(True, 0.05)
            except queue.Empty:
                continue
            # make Task object from request
            task = Task.from_proto(task_proto)
            try:
                res = functools.partial(task.func, *(task, *task.args) if task.bind else task.args, **task.kwargs)()
                task.on_success(res)
            except Retry as e:
                task.error = str(e)
                task.exc = traceback.format_stack()
                if task.on_retry(e):
                    continue
                task.on_fail()
                self.pool.remove_worker(self.ident)
                exc = MaxRetriesExceedException()
                break
            except Exception as e:
                task.error = str(e)
                task.exc = traceback.format_exc()
                task.on_fail()
                self.pool.remove_worker(self.ident)
                exc = e
                break
        if exc is not None:
            raise exc
        self.logger.info("stop")

    def stop(self):
        """Stop worker. multiprocessing or threading part

        """
        pass

    def is_alive(self):
        """Check worker is alive. multiprocessing or threading part

        Returns:
            bool: True if worker is alive or False
        """
        pass


class ProcessWorker(multiprocessing.Process, Worker):
    """See `gromozeka.concurrency.Worker` documentation

    Attributes:
        _stop_event(multiprocessing.Event): Stop event
    """
    __doc__ = Worker.__doc__

    def __init__(self, pool):
        multiprocessing.Process.__init__(self)
        Worker.__init__(self, pool=pool)
        self._stop_event = multiprocessing.Event()

    def run(self):
        Worker.run(self)

    def stop(self, timeout=None):
        self._stop_event.set()


class ThreadWorker(threading.Thread, Worker):
    """See `gromozeka.concurrency.Worker` documentation

    Attributes:
        _stop_event(threading.Event): Stop event
    """
    __doc__ = Worker.__doc__

    def __init__(self, pool):
        threading.Thread.__init__(self)
        Worker.__init__(self, pool=pool)
        self._stop_event = threading.Event()

    def run(self):
        Worker.run(self)

    def stop(self, timeout=None):
        self._stop_event.set()



from gromozeka.concurrency.pool import Pool
from gromozeka.concurrency.scheduler import Scheduler
from gromozeka.concurrency.worker import ThreadWorker, ProcessWorker, Worker

__all__ = ['Pool', 'ThreadWorker', 'ProcessWorker', 'Scheduler', 'Worker']



from gromozeka.primitives.base import Task, RegistryTask, task
from gromozeka.primitives.task_pb2 import Task as ProtoTask, ReplyToBrokerPoint as ProtoReplyToBrokerPoint

__all__ = ['Task', 'RegistryTask', 'task', 'ProtoTask', 'ProtoReplyToBrokerPoint']



import logging
import os
from logging import config


def get_logging():
    from gromozeka.app import get_app
    app_id = get_app().config.app_id
    return {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'console': {
                'format': '%(asctime)-15s %(levelname)s %(name)s [%(threadName)s %(processName)s] %(message)s'
            },
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'console'
            },
        },
        'loggers': {
            '%s' % app_id: {'handlers': ['console'], 'level': 'INFO'},
            '%s.pool' % app_id: {},
            '%s.pool.worker' % app_id: {},
            '%s.pool.worker.*' % app_id: {},
            '%s.scheduler' % app_id: {},
            '%s.broker' % app_id: {},
            '%s.broker.consumer' % app_id: {},
            '%s.broker.consumer.*' % app_id: {},
        }
    }


class Config:
    """Configuration class
    
    Attributes:
        debug(:obj:`bool`): Set logger level. Default: `False`
        broker_reconnect_max_retries(:obj:`int`): Broker reconnect attempts. Default: 10
        broker_reconnect_retry_countdown(:obj:`int`): Sleep time between reconnects (seconds). Default: 10
        broker_url(:obj:`str`): Broker url. Default: 'amqp://guest:guest@localhost:5672/%2F'
        backend_url(:obj:`str`): Backend url. Default: 'redis://localhost'
        backend_reconnect_max_retries(:obj:`int`): Backend reconnect attempts. Default: 10
        backend_reconnect_retry_countdown(:obj:`int`): Sleep time between reconnects (seconds). Default: 10
        backend_results_timelife(:obj:`int`): Results expiration time life (seconds). Default: 3600
    """

    def __init__(self):
        self.debug = False
        self.app_id = 'gromozeka'
        self.broker_url = 'amqp://guest:guest@localhost:5672/%2F'
        self.broker_reconnect_max_retries = 10
        self.broker_reconnect_retry_countdown = 10
        self.backend_url = 'redis://localhost'
        self.backend_reconnect_max_retries = 10
        self.backend_reconnect_retry_countdown = 10
        self.backend_results_timelife = 3600

    def from_env(self):
        """Configure Gromozeka with environment variables

        Returns:
            gromozeka.Config: Configured object
        """

        for name in dir(self):
            if name.startswith('__') or callable(getattr(self, name)):
                continue
            new_attr = os.getenv(name, getattr(self, name))
            type_ = type(getattr(self, name))
            self._set_var_type(type_, name, new_attr)
        self._set_up_logging()
        return self

    def from_dict(self, config_dict):
        """Configure Gromozeka with :obj:`dict`

        Args:
            config_dict(dict): config :obj:`dict`

        Returns:
            gromozeka.Config: Configured object
        """
        for name, value in config_dict.items():
            if not hasattr(self, name):
                from gromozeka import GromozekaException
                raise GromozekaException('Unknown config option `{}`'.format(name))
            type_ = type(getattr(self, name))
            self._set_var_type(type_, name, value)
        self._set_up_logging()
        return self

    def _set_up_logging(self):
        config.dictConfig(get_logging())
        if self.debug:
            logging.getLogger(self.app_id).setLevel(logging.DEBUG)

    def _set_var_type(self, type_, name, attr):
        if type(None) == type_:
            setattr(self, name, attr)
        if type(False) == type_:
            setattr(self, name, False if attr in ('0', 0, 'false', 'False', '') else True)
        else:
            setattr(self, name, type_(attr))


__all__ = ['Config']



import sys
import traceback


class GromozekaException(Exception):
    """Base exception"""
    pass


class SerializationError(GromozekaException):
    pass


class Retry(GromozekaException):
    """Exception for retry.

    Attributes:
        max_retries(:obj:`int`): Maximum number of retries
        retry_countdown(:obj:`float`): Pause between retries (seconds)

    Args:
        *args: Exception
        max_retries(int): Maximum number of retries
        retry_countdown(float): Pause between retries (seconds)

    Use this exceptions in task to retry it

    Examples:
        .. code-block:: python

            @task()
            def some_task():
                try:
                    a = 1 / 0
                except ZeroDivisionError as e:
                    raise Retry(e)

        or with custom number of retries and retry_countdown:


        .. code-block:: python

            @task()
            def some_task():
                try:
                    a = 1 / 0
                except ZeroDivisionError as e:
                    raise Retry(e,max_retries=10, retry_countdown=3)
    """

    def __init__(self, *args, max_retries=None, retry_countdown=None):
        _, _, exc_traceback = sys.exc_info()
        self.traceback = traceback.format_exc() if exc_traceback else ""
        self.max_retries = max_retries
        self.retry_countdown = retry_countdown
        super().__init__(*args)


class MaxRetriesExceedException(GromozekaException):
    """Exception will raise if maximum number of retries is reached

    """
    pass


class BackendException(GromozekaException):
    pass


__all__ = ['GromozekaException', 'Retry', 'MaxRetriesExceedException', 'BackendException', 'SerializationError']



from gromozeka.app import Gromozeka, get_app
from gromozeka.concurrency import ProcessWorker, ThreadWorker
from gromozeka.exceptions import Retry, MaxRetriesExceedException, GromozekaException
from gromozeka.primitives import task, ProtoTask, ProtoReplyToBrokerPoint
from gromozeka.primitives.base import BrokerPoint, TaskDeserializator

__all__ = [
    'Gromozeka',
    'get_app',
    'Retry',
    'MaxRetriesExceedException',
    'GromozekaException',
    'ProcessWorker',
    'ThreadWorker',
    'task',
    'BrokerPoint',
    'ProtoTask',
    'ProtoReplyToBrokerPoint',
    'TaskDeserializator'
]

__version__ = '0.1.3'



"""Commands to synchronize processes or threads

"""


class Command:
    """Helper class to make commands

    """
    # command number. Find it in `gromozeka.primitives.commands`
    command = None
    # command arguments
    args = None

    def __init__(self, command, args=None):
        self.command = command
        self.args = args

    def as_dict(self):
        return {'command': self.command, 'args': self.args}

    def as_tuple(self):
        return self.command, self.args


POOL_STOP = 1
POOL_GROW = 2
POOL_SHRINK = 3
POOL_SIZE = 4
POOL_REMOVE_WORKER = 5
BROKER_TASK_REGISTER = 6
BROKER_TASK_SEND = 7
BROKER_TASK_SEND_DELAYED = 8
BROKER_TASK_DONE = 9
BROKER_TASK_REJECT = 10
BROKER_ON_POOL_SIZE_CHANGED = 11
BACKEND_RESULT_GET = 12
BACKEND_RESULT_SET = 13
BACKEND_RESULT_DEL = 14
BACKEND_RESULTS_DEL = 15
BACKEND_GRAPH_INIT = 16
BACKEND_GRAPH_GET = 17
BACKEND_GRAPH_UPDATE = 18
BACKEND_GRAPH_RESULT_SET = 19
BACKEND_IS_GROUP_COMPLETED = 20
BACKEND_GROUP_ADD_RESULT = 21
BACKEND_GROUP_GET_RESULT = 22
BACKEND_CHAIN_GET_RESULT = 23


# POOL
def pool_stop():
    """Command to stop pool

    Returns:
        Command:
    """
    return Command(command=POOL_STOP, args=None).as_tuple()


def pool_grow(n):
    """Command to grow pool

    Args:
        n: Number of workers to grow

    Returns:
        Command:
    """
    return Command(command=POOL_GROW, args=n).as_tuple()


def pool_shrink(n):
    """Command to shrink pool

    Args:
        n: Number of workers to shrink

    Returns:
        Command:
    """
    return Command(command=POOL_SHRINK, args=n).as_tuple()


def pool_size():
    """Command to get pool size

    Returns:
        Command:
    """
    return Command(command=POOL_SIZE, args=None).as_tuple()


def pool_remove_worker(worker_ident):
    """Command to remove worker from pool

    Args:
        worker_ident(int): Worker identification

    Returns:
        Command:
    """
    return Command(command=POOL_REMOVE_WORKER, args=worker_ident).as_tuple()


# BROKER
def broker_task_register(task_id, broker_point, options, deserializator):
    """Command to add new consumer with `broker_point` to task

    Args:
        task_id(str): Unique task identification
        broker_point(primitives.BrokerPointType): Broker entry
        options: Specific broker options. See NatsOptions broker for example

    Returns:
        Command:
    """
    return Command(command=BROKER_TASK_REGISTER,
                   args={'task_id': task_id, 'broker_point': broker_point, 'options': options,
                         'deserializator': deserializator}).as_tuple()


def broker_task_send(task_uuid, request, broker_point, reply_to=None):
    """Command to publish `primitives.Request` to customer

    Args:
        task_uuid(str): task identification
        request: Serialized request
        broker_point(gromozeka.BrokerPoint):
        reply_to(gromozeka.BrokerPoint):

    Returns:
        Command:
    """
    return Command(command=BROKER_TASK_SEND,
                   args={'task_uuid': task_uuid, 'request': request, 'broker_point': broker_point,
                         'reply_to': reply_to}).as_tuple()


def broker_task_send_delayed(task_uuid, request, broker_point, delay):
    """Command to publish `primitives.Request` to customer

    Args:
        task_uuid(str): task identification
        request(primitives.Request): Request to publish
        broker_point:
        delay:

    Returns:
        Command:
    """
    return Command(command=BROKER_TASK_SEND_DELAYED,
                   args={'task_uuid': task_uuid, 'request': request, 'broker_point': broker_point,
                         'delay': delay}).as_tuple()


def broker_task_done(task_uuid, broker_point, delivery_tag):
    """Command to acknowledge task message by it's broker_adapter point, delivery_tag or scheduler_tag

    Args:
        task_uuid(str): task identification
        broker_point(primitives.BrokerPointType): Broker entry
        delivery_tag(int): Task delivery tag
    Returns:
        Command:
    """
    return Command(command=BROKER_TASK_DONE,
                   args={'task_uuid': task_uuid, 'broker_point': broker_point, 'delivery_tag': delivery_tag}).as_tuple()


def broker_task_reject(task_uuid, broker_point, delivery_tag):
    """Command to reject task message by it's broker_adapter point, delivery_tag or scheduler_tag

    Args:
        task_uuid(str): task identification
        broker_point(primitives.BrokerPointType): Broker entry
        delivery_tag(int): Task delivery tag

    Returns:
        Command:
    """
    return Command(command=BROKER_TASK_REJECT,
                   args={'task_uuid': task_uuid, 'broker_point': broker_point, 'delivery_tag': delivery_tag}).as_tuple()


def broker_on_pool_size_changed():
    """Command to change prefetch_count

    Returns:
        Command:
    """
    return Command(command=BROKER_ON_POOL_SIZE_CHANGED, args=None).as_tuple()


def backend_result_get(task_uuid, graph_uuid):
    """Command to get result from backend_adapter

    Args:
        task_uuid: Task uuid
        graph_uuid: Graph uuid

    Returns:
        Command:
    """
    return Command(command=BACKEND_RESULT_GET, args={'task_uuid': task_uuid, 'graph_uuid': graph_uuid}).as_tuple()


def backend_result_set(task_uuid, result, graph_uuid):
    """Command to set task result

    Args:
        task_uuid: Task uuid
        result: Result
        graph_uuid: Graph uuid
    Returns:
        Command:
    """
    return Command(command=BACKEND_RESULT_SET,
                   args={'task_uuid': task_uuid, 'result': result, 'graph_uuid': graph_uuid}).as_tuple()


def backend_result_del(task_uuid):
    """Command to delete result

    Args:
        task_uuid: Task uuid

    Returns:
        Command:
    """
    return Command(command=BACKEND_RESULT_DEL, args=task_uuid).as_tuple()


def backend_results_del(*task_uuids):
    """Command to delete results

    Args:
        task_uuids: Task uuids

    Returns:
        Command:
    """
    return Command(command=BACKEND_RESULTS_DEL, args=task_uuids).as_tuple()


def backend_graph_init(graph_dict):
    return Command(command=BACKEND_GRAPH_INIT, args=graph_dict).as_tuple()


def backend_graph_get(graph_uuid):
    return Command(command=BACKEND_GRAPH_GET, args=graph_uuid).as_tuple()


def backend_graph_update(graph_uuid, verticies, graph_state=None, error_task_uuid=None, short_error=None):
    return Command(command=BACKEND_GRAPH_UPDATE,
                   args={'graph_uuid': graph_uuid, 'verticies': verticies, 'graph_state': graph_state,
                         'error_task_uuid': error_task_uuid, 'short_error': short_error}).as_tuple()


def backend_graph_result_set(graph_uuid, task_uuid, result):
    return Command(command=BACKEND_GRAPH_RESULT_SET,
                   args={'graph_uuid': graph_uuid, 'task_uuid': task_uuid, 'result': result}).as_tuple()


def backend_is_group_completed(graph_uuid, group_uuid, expected):
    return Command(command=BACKEND_IS_GROUP_COMPLETED,
                   args={'graph_uuid': graph_uuid, 'group_uuid': group_uuid, 'expected': expected}).as_tuple()


def backend_group_add_result(graph_uuid, group_uuid, task_uuid, result):
    return Command(command=BACKEND_GROUP_ADD_RESULT,
                   args={'graph_uuid': graph_uuid, 'group_uuid': group_uuid, 'task_uuid': task_uuid,
                         'result': result}).as_tuple()


def backend_group_get_result(graph_uuid, group_uuid):
    return Command(command=BACKEND_GROUP_GET_RESULT,
                   args={'graph_uuid': graph_uuid, 'group_uuid': group_uuid}).as_tuple()


def backend_chain_get_result(graph_uuid, chain_uuid):
    return Command(command=BACKEND_CHAIN_GET_RESULT,
                   args={'graph_uuid': graph_uuid, 'chain_uuid': chain_uuid}).as_tuple()



import io
import os
import re
from os.path import join, dirname

from setuptools import setup, find_packages

PACKAGE = "gromozeka"
NAME = "gromozeka"
DESCRIPTION = "Distributed task queue."
AUTHOR = "Klimov Konstantin"
AUTHOR_EMAIL = "moelius1983@gmail.com"
URL = "https://github.com/moelius/gromozeka"

with io.open('gromozeka/__init__.py', 'rt', encoding='utf8') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)


def _strip_comments(l):
    return l.split('#', 1)[0].strip()


def _pip_requirement(req):
    if req.startswith('-r '):
        _, path = req.split()
        return requirements_list(*path.split('/'))
    return [req]


def _requirements_list(*files):
    return [
        _pip_requirement(r) for r in
        (_strip_comments(l) for l in open(os.path.join(os.getcwd(), 'requirements', *files)).readlines()) if r]


def requirements_list(*files):
    return [req for sub in _requirements_list(*files) for req in sub]


setup(
    name=NAME,
    version=version,
    description=DESCRIPTION,
    long_description=open(join(dirname(__file__), 'README.rst')).read(),
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    keywords='task job queue distributed messaging pika rabbit rabbitmq pool worker',
    license="MIT",
    url=URL,
    packages=find_packages(exclude=["tests.*", "tests"]),
    install_requires=requirements_list('base.txt'),
    extras_require={
        'docs': requirements_list('docs.txt'),
        'visual': requirements_list('visual.txt'),
    },
    zip_safe=False,
)



import logging
import queue

from gromozeka.concurrency import Pool, commands
from gromozeka.concurrency import ThreadWorker
from gromozeka.exceptions import GromozekaException


def get_broker_factory(app):
    """

    Args:
        app (gromozeka.Gromozeka):

    Returns:
        gromozeka.brokers.base.BrokerAdapter:
    """
    if app.config.broker_url.startswith('amqp://'):
        from gromozeka.brokers import RabbitMQPikaAdaptee
        broker_adapter = BrokerAdapter(RabbitMQPikaAdaptee(app=app))
    else:
        raise GromozekaException('unknown broker connection url')
    broker_adapter.broker.configure()
    return broker_adapter


class BrokerInterface:
    def __init__(self, app):
        """

        Args:
            app(gromozeka.Gromozeka):
        """
        self.logger = logging.getLogger("%s.broker" % app.config.app_id)
        self.app = app

    @staticmethod
    def worker_run(self):
        """This method will be in thread - start to work here

        """
        raise NotImplementedError

    def stop(self):
        """Stop broker

        """
        raise NotImplementedError

    def configure(self):
        """Configure broker

        """
        raise NotImplementedError

    def task_register(self, broker_point, task_id, options, deserializator):
        """

        Args:
            broker_point(gromozeka.primitives.protocol.ProtoBrokerPoint): Broker entry
            task_id(str): Task identification
            options: Specific broker options.
            deserializator(gromozeka.primitives.base.TaskDeserializator):

        """
        raise NotImplementedError

    def task_send(self, task_uuid, request, broker_point, reply_to=None):
        """

        Args:
            task_uuid(str): task identification
            request: serialized request
            broker_point(gromozeka.BrokerPoint):
            reply_to(gromozeka.BrokerPoint):

        """
        raise NotImplementedError

    def task_send_delayed(self, task_uuid, request, broker_point, delay):
        """

        Args:
            task_uuid(str): task identification
            request: serialized request
            broker_point(gromozeka.BrokerPoint):
            delay:

        Returns:

        """
        raise NotImplementedError

    def task_done(self, task_uuid, broker_point, delivery_tag):
        """

        Args:
            task_uuid(str): task identification
            broker_point(gromozeka.primitives.protocol.ProtoBrokerPoint): Broker entry
            delivery_tag: Broker delivery tag (unique message identification)

        """
        raise NotImplementedError

    def task_reject(self, task_uuid, broker_point, delivery_tag):
        """

        Args:
            task_uuid(str): task identification
            broker_point(gromozeka.primitives.protocol.ProtoBrokerPoint): Broker entry
            delivery_tag: Broker delivery tag (unique message identification)

        """
        raise NotImplementedError

    def on_pool_size_changed(self):
        """This method will be run, when worker pool size change

        """
        raise NotImplementedError

    def wait_for_start(self):
        """Not required method for check broker is started

        Returns:

        """
        pass


class BrokerAdapter(Pool):
    def __init__(self, broker):
        """

        Args:
            broker(BrokerInterface):
        """
        self.logger = logging.getLogger("%s.broker" % broker.app.config.app_id)
        self.broker = broker
        self.worker = type('BrokerWorker', (ThreadWorker,),
                           {'app': self.broker.app, 'broker': self.broker, 'run': self.broker.worker_run,
                            'logger': self.logger})
        super().__init__(max_workers=1, worker_class=self.worker, logger=self.logger)

    def listen_cmd(self):
        """Listen to commands

        """
        while not self._stop_event.is_set():
            try:
                command, args = self.cmd.get(0.05)
            except queue.Empty:
                continue
            if command == commands.POOL_GROW:
                super()._grow(args)
            elif command == commands.POOL_STOP:
                self.broker.stop()
                super().stop_()
            elif command == commands.BROKER_TASK_REGISTER:
                task_id, broker_point, options = args['task_id'], args['broker_point'], args['options']
                deserializator = args['deserializator']
                self.broker.task_register(task_id=task_id, broker_point=broker_point, options=options,
                                          deserializator=deserializator)
            elif command == commands.BROKER_TASK_SEND:
                task_uuid, request = args['task_uuid'], args['request']
                broker_point, reply_to = args['broker_point'], args['reply_to']
                self.broker.task_send(task_uuid=task_uuid, request=request, broker_point=broker_point,
                                      reply_to=reply_to)
            elif command == commands.BROKER_TASK_SEND_DELAYED:
                task_uuid, request = args['task_uuid'], args['request']
                broker_point, delay = args['broker_point'], args['delay']
                self.broker.task_send_delayed(task_uuid=task_uuid, request=request, broker_point=broker_point,
                                              delay=delay)
            elif command == commands.BROKER_TASK_DONE:
                task_uuid, broker_point, delivery_tag = args['task_uuid'], args['broker_point'], args['delivery_tag']
                self.broker.task_done(task_uuid=task_uuid, broker_point=broker_point, delivery_tag=delivery_tag)
            elif command == commands.BROKER_TASK_REJECT:
                task_uuid, broker_point = args['task_uuid'], args['broker_point']
                delivery_tag = args['delivery_tag']
                self.broker.task_reject(task_uuid=task_uuid, broker_point=broker_point, delivery_tag=delivery_tag)
            elif command == commands.BROKER_ON_POOL_SIZE_CHANGED:
                self.broker.on_pool_size_changed()

    def start(self):
        super().start()
        self.broker.wait_for_start()

    def task_register(self, broker_point, task_id, options=None, deserializator=None):
        """Register task in broker

        Args:
            broker_point(gromozeka.primitives.protocol.ProtoBrokerPoint): Broker entry
            task_id(str): Task identification
            options: Specific broker options. See NatsOptions broker for example

        """
        self.cmd.put(commands.broker_task_register(task_id=task_id, broker_point=broker_point, options=options,
                                                   deserializator=deserializator))

    def task_send(self, task_uuid, request, broker_point, reply_to=None):
        """

        Args:
            task_uuid(str): task identification
            request: serialized request
            broker_point(gromozeka.BrokerPoint):

        """
        self.cmd.put(commands.broker_task_send(task_uuid=task_uuid, request=request, broker_point=broker_point,
                                               reply_to=reply_to))

    def task_send_delayed(self, task_uuid, request, broker_point, delay):
        """

        Args:
            task_uuid(str): task identification
            request(gromozeka.primitives.protocol.Request): Request object
            broker_point:
            delay:

        """
        self.cmd.put(commands.broker_task_send_delayed(task_uuid=task_uuid, request=request, broker_point=broker_point,
                                                       delay=delay))

    def task_done(self, task_uuid, broker_point, delivery_tag):
        """

        Args:
            task_uuid(str): task identification
            broker_point(gromozeka.primitives.protocol.ProtoBrokerPoint): Broker entry
            delivery_tag: Broker delivery tag (unique message identification)

        """
        self.cmd.put(
            commands.broker_task_done(task_uuid=task_uuid, broker_point=broker_point,
                                      delivery_tag=delivery_tag))

    def task_reject(self, task_uuid, broker_point, delivery_tag):
        """


        Args:
            task_uuid(str): task identification
            broker_point(gromozeka.primitives.protocol.ProtoBrokerPoint): Broker entry
            delivery_tag: Broker delivery tag (unique message identification)

        """
        self.cmd.put(
            commands.broker_task_reject(task_uuid=task_uuid, broker_point=broker_point, delivery_tag=delivery_tag))

    def on_pool_size_changed(self):
        """This method will be run, when worker pool size change

        """
        self.cmd.put(commands.broker_on_pool_size_changed())
